import functools
from typing import (
    Any,
    Callable,
    Protocol,
    Type,
    TypeVar,
    Union,
    cast,
    get_type_hints,
    runtime_checkable,
)

from pydantic import BaseModel

from hatchet_sdk.v0 import ConcurrencyLimitStrategy
from hatchet_sdk.v0.contracts.workflows_pb2 import (
    CreateWorkflowJobOpts,
    CreateWorkflowStepOpts,
    CreateWorkflowVersionOpts,
    StickyStrategy,
    WorkflowConcurrencyOpts,
    WorkflowKind,
)
from hatchet_sdk.v0.logger import logger
from hatchet_sdk.v0.utils.typing import is_basemodel_subclass


class WorkflowStepProtocol(Protocol):
    def __call__(self, *args: Any, **kwargs: Any) -> Any: ...

    __name__: str

    _step_name: str
    _step_timeout: str | None
    _step_parents: list[str]
    _step_retries: int | None
    _step_rate_limits: list[str] | None
    _step_desired_worker_labels: dict[str, str]
    _step_backoff_factor: float | None
    _step_backoff_max_seconds: int | None

    _concurrency_fn_name: str
    _concurrency_max_runs: int | None
    _concurrency_limit_strategy: str | None

    _on_failure_step_name: str
    _on_failure_step_timeout: str | None
    _on_failure_step_retries: int
    _on_failure_step_rate_limits: list[str] | None
    _on_failure_step_backoff_factor: float | None
    _on_failure_step_backoff_max_seconds: int | None


StepsType = list[tuple[str, WorkflowStepProtocol]]

T = TypeVar("T")
TW = TypeVar("TW", bound="WorkflowInterface")


class ConcurrencyExpression:
    """
    Defines concurrency limits for a workflow using a CEL expression.

    Args:
        expression (str): CEL expression to determine concurrency grouping. (i.e. "input.user_id")
        max_runs (int): Maximum number of concurrent workflow runs.
        limit_strategy (ConcurrencyLimitStrategy): Strategy for handling limit violations.

    Example:
        ConcurrencyExpression("input.user_id", 5, ConcurrencyLimitStrategy.CANCEL_IN_PROGRESS)
    """

    def __init__(
        self, expression: str, max_runs: int, limit_strategy: ConcurrencyLimitStrategy
    ):
        self.expression = expression
        self.max_runs = max_runs
        self.limit_strategy = limit_strategy


@runtime_checkable
class WorkflowInterface(Protocol):
    def get_name(self, namespace: str) -> str: ...

    def get_actions(self, namespace: str) -> list[tuple[str, Callable[..., Any]]]: ...

    def get_create_opts(self, namespace: str) -> Any: ...

    on_events: list[str] | None
    on_crons: list[str] | None
    name: str
    version: str
    timeout: str
    schedule_timeout: str
    sticky: Union[StickyStrategy.Value, None]  # type: ignore[name-defined]
    default_priority: int | None
    concurrency_expression: ConcurrencyExpression | None
    input_validator: Type[BaseModel] | None


class WorkflowMeta(type):
    def __new__(
        cls: Type["WorkflowMeta"],
        name: str,
        bases: tuple[type, ...],
        attrs: dict[str, Any],
    ) -> "WorkflowMeta":
        def _create_steps_actions_list(name: str) -> StepsType:
            return [
                (getattr(func, name), attrs.pop(func_name))
                for func_name, func in list(attrs.items())
                if hasattr(func, name)
            ]

        concurrencyActions = _create_steps_actions_list("_concurrency_fn_name")
        steps = _create_steps_actions_list("_step_name")

        onFailureSteps = _create_steps_actions_list("_on_failure_step_name")

        # Define __init__ and get_step_order methods
        original_init = attrs.get("__init__")  # Get the original __init__ if it exists

        def __init__(self: TW, *args: Any, **kwargs: Any) -> None:
            if original_init:
                original_init(self, *args, **kwargs)  # Call original __init__

        def get_service_name(namespace: str) -> str:
            return f"{namespace}{name.lower()}"

        @functools.cache
        def get_actions(self: TW, namespace: str) -> StepsType:
            serviceName = get_service_name(namespace)

            func_actions = [
                (serviceName + ":" + func_name, func) for func_name, func in steps
            ]
            concurrency_actions = [
                (serviceName + ":" + func_name, func)
                for func_name, func in concurrencyActions
            ]
            onFailure_actions = [
                (serviceName + ":" + func_name, func)
                for func_name, func in onFailureSteps
            ]

            return func_actions + concurrency_actions + onFailure_actions

        # Add these methods and steps to class attributes
        attrs["__init__"] = __init__
        attrs["get_actions"] = get_actions

        for step_name, step_func in steps:
            attrs[step_name] = step_func

        def get_name(self: TW, namespace: str) -> str:
            return namespace + cast(str, attrs["name"])

        attrs["get_name"] = get_name

        cron_triggers = attrs["on_crons"]
        version = attrs["version"]
        schedule_timeout = attrs["schedule_timeout"]
        sticky = attrs["sticky"]
        default_priority = attrs["default_priority"]

        @functools.cache
        def get_create_opts(self: TW, namespace: str) -> CreateWorkflowVersionOpts:
            serviceName = get_service_name(namespace)
            name = self.get_name(namespace)
            event_triggers = [namespace + event for event in attrs["on_events"]]
            createStepOpts: list[CreateWorkflowStepOpts] = [
                CreateWorkflowStepOpts(
                    readable_id=step_name,
                    action=serviceName + ":" + step_name,
                    timeout=func._step_timeout or "60s",
                    inputs="{}",
                    parents=[x for x in func._step_parents],
                    retries=func._step_retries,
                    rate_limits=func._step_rate_limits,  # type: ignore[arg-type]
                    worker_labels=func._step_desired_worker_labels,  # type: ignore[arg-type]
                    backoff_factor=func._step_backoff_factor,
                    backoff_max_seconds=func._step_backoff_max_seconds,
                )
                for step_name, func in steps
            ]

            concurrency: WorkflowConcurrencyOpts | None = None

            if len(concurrencyActions) > 0:
                action = concurrencyActions[0]

                concurrency = WorkflowConcurrencyOpts(
                    action=serviceName + ":" + action[0],
                    max_runs=action[1]._concurrency_max_runs,
                    limit_strategy=action[1]._concurrency_limit_strategy,
                )

            if self.concurrency_expression:
                concurrency = WorkflowConcurrencyOpts(
                    expression=self.concurrency_expression.expression,
                    max_runs=self.concurrency_expression.max_runs,
                    limit_strategy=self.concurrency_expression.limit_strategy,
                )

            if len(concurrencyActions) > 0 and self.concurrency_expression:
                raise ValueError(
                    "Error: Both concurrencyActions and concurrency_expression are defined. Please use only one concurrency configuration method."
                )

            on_failure_job: CreateWorkflowJobOpts | None = None

            if len(onFailureSteps) > 0:
                func_name, func = onFailureSteps[0]
                on_failure_job = CreateWorkflowJobOpts(
                    name=name + "-on-failure",
                    steps=[
                        CreateWorkflowStepOpts(
                            readable_id=func_name,
                            action=serviceName + ":" + func_name,
                            timeout=func._on_failure_step_timeout or "60s",
                            inputs="{}",
                            parents=[],
                            retries=func._on_failure_step_retries,
                            rate_limits=func._on_failure_step_rate_limits,  # type: ignore[arg-type]
                            backoff_factor=func._on_failure_step_backoff_factor,
                            backoff_max_seconds=func._on_failure_step_backoff_max_seconds,
                        )
                    ],
                )

            validated_priority = (
                max(1, min(3, default_priority)) if default_priority else None
            )
            if validated_priority != default_priority:
                logger.warning(
                    "Warning: Default Priority Must be between 1 and 3 -- inclusively. Adjusted to be within the range."
                )

            return CreateWorkflowVersionOpts(
                name=name,
                kind=WorkflowKind.DAG,
                version=version,
                event_triggers=event_triggers,
                cron_triggers=cron_triggers,
                schedule_timeout=schedule_timeout,
                sticky=sticky,
                jobs=[
                    CreateWorkflowJobOpts(
                        name=name,
                        steps=createStepOpts,
                    )
                ],
                on_failure_job=on_failure_job,
                concurrency=concurrency,
                default_priority=validated_priority,
            )

        attrs["get_create_opts"] = get_create_opts

        return super(WorkflowMeta, cls).__new__(cls, name, bases, attrs)



# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc
import warnings

from . import workflows_pb2 as workflows__pb2

GRPC_GENERATED_VERSION = '1.64.1'
GRPC_VERSION = grpc.__version__
EXPECTED_ERROR_RELEASE = '1.65.0'
SCHEDULED_RELEASE_DATE = 'June 25, 2024'
_version_not_supported = False

try:
    from grpc._utilities import first_version_is_lower
    _version_not_supported = first_version_is_lower(GRPC_VERSION, GRPC_GENERATED_VERSION)
except ImportError:
    _version_not_supported = True

if _version_not_supported:
    warnings.warn(
        f'The grpc package installed is at version {GRPC_VERSION},'
        + f' but the generated code in workflows_pb2_grpc.py depends on'
        + f' grpcio>={GRPC_GENERATED_VERSION}.'
        + f' Please upgrade your grpc module to grpcio>={GRPC_GENERATED_VERSION}'
        + f' or downgrade your generated code using grpcio-tools<={GRPC_VERSION}.'
        + f' This warning will become an error in {EXPECTED_ERROR_RELEASE},'
        + f' scheduled for release on {SCHEDULED_RELEASE_DATE}.',
        RuntimeWarning
    )


class WorkflowServiceStub(object):
    """WorkflowService represents a set of RPCs for managing workflows.
    """

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.PutWorkflow = channel.unary_unary(
                '/WorkflowService/PutWorkflow',
                request_serializer=workflows__pb2.PutWorkflowRequest.SerializeToString,
                response_deserializer=workflows__pb2.WorkflowVersion.FromString,
                _registered_method=True)
        self.ScheduleWorkflow = channel.unary_unary(
                '/WorkflowService/ScheduleWorkflow',
                request_serializer=workflows__pb2.ScheduleWorkflowRequest.SerializeToString,
                response_deserializer=workflows__pb2.WorkflowVersion.FromString,
                _registered_method=True)
        self.TriggerWorkflow = channel.unary_unary(
                '/WorkflowService/TriggerWorkflow',
                request_serializer=workflows__pb2.TriggerWorkflowRequest.SerializeToString,
                response_deserializer=workflows__pb2.TriggerWorkflowResponse.FromString,
                _registered_method=True)
        self.BulkTriggerWorkflow = channel.unary_unary(
                '/WorkflowService/BulkTriggerWorkflow',
                request_serializer=workflows__pb2.BulkTriggerWorkflowRequest.SerializeToString,
                response_deserializer=workflows__pb2.BulkTriggerWorkflowResponse.FromString,
                _registered_method=True)
        self.PutRateLimit = channel.unary_unary(
                '/WorkflowService/PutRateLimit',
                request_serializer=workflows__pb2.PutRateLimitRequest.SerializeToString,
                response_deserializer=workflows__pb2.PutRateLimitResponse.FromString,
                _registered_method=True)


class WorkflowServiceServicer(object):
    """WorkflowService represents a set of RPCs for managing workflows.
    """

    def PutWorkflow(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def ScheduleWorkflow(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def TriggerWorkflow(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def BulkTriggerWorkflow(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def PutRateLimit(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_WorkflowServiceServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'PutWorkflow': grpc.unary_unary_rpc_method_handler(
                    servicer.PutWorkflow,
                    request_deserializer=workflows__pb2.PutWorkflowRequest.FromString,
                    response_serializer=workflows__pb2.WorkflowVersion.SerializeToString,
            ),
            'ScheduleWorkflow': grpc.unary_unary_rpc_method_handler(
                    servicer.ScheduleWorkflow,
                    request_deserializer=workflows__pb2.ScheduleWorkflowRequest.FromString,
                    response_serializer=workflows__pb2.WorkflowVersion.SerializeToString,
            ),
            'TriggerWorkflow': grpc.unary_unary_rpc_method_handler(
                    servicer.TriggerWorkflow,
                    request_deserializer=workflows__pb2.TriggerWorkflowRequest.FromString,
                    response_serializer=workflows__pb2.TriggerWorkflowResponse.SerializeToString,
            ),
            'BulkTriggerWorkflow': grpc.unary_unary_rpc_method_handler(
                    servicer.BulkTriggerWorkflow,
                    request_deserializer=workflows__pb2.BulkTriggerWorkflowRequest.FromString,
                    response_serializer=workflows__pb2.BulkTriggerWorkflowResponse.SerializeToString,
            ),
            'PutRateLimit': grpc.unary_unary_rpc_method_handler(
                    servicer.PutRateLimit,
                    request_deserializer=workflows__pb2.PutRateLimitRequest.FromString,
                    response_serializer=workflows__pb2.PutRateLimitResponse.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'WorkflowService', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))
    server.add_registered_method_handlers('WorkflowService', rpc_method_handlers)


 # This class is part of an EXPERIMENTAL API.
class WorkflowService(object):
    """WorkflowService represents a set of RPCs for managing workflows.
    """

    @staticmethod
    def PutWorkflow(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/WorkflowService/PutWorkflow',
            workflows__pb2.PutWorkflowRequest.SerializeToString,
            workflows__pb2.WorkflowVersion.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def ScheduleWorkflow(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/WorkflowService/ScheduleWorkflow',
            workflows__pb2.ScheduleWorkflowRequest.SerializeToString,
            workflows__pb2.WorkflowVersion.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def TriggerWorkflow(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/WorkflowService/TriggerWorkflow',
            workflows__pb2.TriggerWorkflowRequest.SerializeToString,
            workflows__pb2.TriggerWorkflowResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def BulkTriggerWorkflow(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/WorkflowService/BulkTriggerWorkflow',
            workflows__pb2.BulkTriggerWorkflowRequest.SerializeToString,
            workflows__pb2.BulkTriggerWorkflowResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def PutRateLimit(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/WorkflowService/PutRateLimit',
            workflows__pb2.PutRateLimitRequest.SerializeToString,
            workflows__pb2.PutRateLimitResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)



from typing import Any, Mapping, Type, TypeGuard

from pydantic import BaseModel


def is_basemodel_subclass(model: Any) -> TypeGuard[Type[BaseModel]]:
    try:
        return issubclass(model, BaseModel)
    except TypeError:
        return False


class TaskIOValidator(BaseModel):
    workflow_input: Type[BaseModel] | None = None
    step_output: Type[BaseModel] | None = None


JSONSerializableMapping = Mapping[str, Any]



import asyncio

from hatchet_sdk.clients.rest.api.worker_api import WorkerApi
from hatchet_sdk.clients.rest.api_client import ApiClient
from hatchet_sdk.clients.rest.models.update_worker_request import UpdateWorkerRequest
from hatchet_sdk.clients.rest.models.worker import Worker
from hatchet_sdk.clients.rest.models.worker_list import WorkerList
from hatchet_sdk.clients.v1.api_client import BaseRestClient


class WorkersClient(BaseRestClient):
    def _wa(self, client: ApiClient) -> WorkerApi:
        return WorkerApi(client)

    def get(self, worker_id: str) -> Worker:
        with self.client() as client:
            return self._wa(client).worker_get(worker_id)

    async def aio_get(self, worker_id: str) -> Worker:
        return await asyncio.to_thread(self.get, worker_id)

    def list(
        self,
    ) -> WorkerList:
        with self.client() as client:
            return self._wa(client).worker_list(
                tenant=self.client_config.tenant_id,
            )

    async def aio_list(
        self,
    ) -> WorkerList:
        return await asyncio.to_thread(self.list)

    def update(self, worker_id: str, opts: UpdateWorkerRequest) -> Worker:
        with self.client() as client:
            return self._wa(client).worker_update(
                worker=worker_id,
                update_worker_request=opts,
            )

    async def aio_update(self, worker_id: str, opts: UpdateWorkerRequest) -> Worker:
        return await asyncio.to_thread(self.update, worker_id, opts)



import json
from collections.abc import AsyncIterator
from typing import Any, Literal, cast

import grpc
import grpc.aio
from pydantic import BaseModel, ConfigDict

from hatchet_sdk.clients.listeners.pooled_listener import PooledListener
from hatchet_sdk.clients.rest.tenacity_utils import tenacity_retry
from hatchet_sdk.connection import new_conn
from hatchet_sdk.contracts.v1.dispatcher_pb2 import (
    DurableEvent,
    ListenForDurableEventRequest,
)
from hatchet_sdk.contracts.v1.dispatcher_pb2 import (
    RegisterDurableEventRequest as RegisterDurableEventRequestProto,
)
from hatchet_sdk.contracts.v1.dispatcher_pb2_grpc import V1DispatcherStub
from hatchet_sdk.contracts.v1.shared.condition_pb2 import DurableEventListenerConditions
from hatchet_sdk.metadata import get_metadata
from hatchet_sdk.waits import SleepCondition, UserEventCondition

DEFAULT_DURABLE_EVENT_LISTENER_RETRY_INTERVAL = 3  # seconds
DEFAULT_DURABLE_EVENT_LISTENER_RETRY_COUNT = 5
DEFAULT_DURABLE_EVENT_LISTENER_INTERRUPT_INTERVAL = 1800  # 30 minutes


class RegisterDurableEventRequest(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True)

    task_id: str
    signal_key: str
    conditions: list[SleepCondition | UserEventCondition]

    def to_proto(self) -> RegisterDurableEventRequestProto:
        return RegisterDurableEventRequestProto(
            task_id=self.task_id,
            signal_key=self.signal_key,
            conditions=DurableEventListenerConditions(
                sleep_conditions=[
                    c.to_proto()
                    for c in self.conditions
                    if isinstance(c, SleepCondition)
                ],
                user_event_conditions=[
                    c.to_proto()
                    for c in self.conditions
                    if isinstance(c, UserEventCondition)
                ],
            ),
        )


class ParsedKey(BaseModel):
    task_id: str
    signal_key: str


class DurableEventListener(
    PooledListener[ListenForDurableEventRequest, DurableEvent, V1DispatcherStub]
):
    def _generate_key(self, task_id: str, signal_key: str) -> str:
        return task_id + ":" + signal_key

    def generate_key(self, response: DurableEvent) -> str:
        return self._generate_key(
            task_id=response.task_id,
            signal_key=response.signal_key,
        )

    def parse_key(self, key: str) -> ParsedKey:
        task_id, signal_key = key.split(":", maxsplit=1)

        return ParsedKey(
            task_id=task_id,
            signal_key=signal_key,
        )

    async def create_subscription(
        self,
        request: AsyncIterator[ListenForDurableEventRequest],
        metadata: tuple[tuple[str, str]],
    ) -> grpc.aio.UnaryStreamCall[ListenForDurableEventRequest, DurableEvent]:
        if self.client is None:
            conn = new_conn(self.config, True)
            self.client = V1DispatcherStub(conn)

        return cast(
            grpc.aio.UnaryStreamCall[ListenForDurableEventRequest, DurableEvent],
            self.client.ListenForDurableEvent(
                request,  # type: ignore[arg-type]
                metadata=metadata,
            ),
        )

    def create_request_body(self, item: str) -> ListenForDurableEventRequest:
        key = self.parse_key(item)
        return ListenForDurableEventRequest(
            task_id=key.task_id,
            signal_key=key.signal_key,
        )

    @tenacity_retry
    def register_durable_event(
        self, request: RegisterDurableEventRequest
    ) -> Literal[True]:
        conn = new_conn(self.config, True)
        client = V1DispatcherStub(conn)

        client.RegisterDurableEvent(
            request.to_proto(),
            timeout=5,
            metadata=get_metadata(self.token),
        )

        return True

    @tenacity_retry
    async def result(self, task_id: str, signal_key: str) -> dict[str, Any]:
        key = self._generate_key(task_id, signal_key)

        event = await self.subscribe(key)

        return cast(dict[str, Any], json.loads(event.data.decode("utf-8")))



from hatchet_sdk.contracts.v1.shared import condition_pb2 as _condition_pb2
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Mapping as _Mapping, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor

class RegisterDurableEventRequest(_message.Message):
    __slots__ = ("task_id", "signal_key", "conditions")
    TASK_ID_FIELD_NUMBER: _ClassVar[int]
    SIGNAL_KEY_FIELD_NUMBER: _ClassVar[int]
    CONDITIONS_FIELD_NUMBER: _ClassVar[int]
    task_id: str
    signal_key: str
    conditions: _condition_pb2.DurableEventListenerConditions
    def __init__(self, task_id: _Optional[str] = ..., signal_key: _Optional[str] = ..., conditions: _Optional[_Union[_condition_pb2.DurableEventListenerConditions, _Mapping]] = ...) -> None: ...

class RegisterDurableEventResponse(_message.Message):
    __slots__ = ()
    def __init__(self) -> None: ...

class ListenForDurableEventRequest(_message.Message):
    __slots__ = ("task_id", "signal_key")
    TASK_ID_FIELD_NUMBER: _ClassVar[int]
    SIGNAL_KEY_FIELD_NUMBER: _ClassVar[int]
    task_id: str
    signal_key: str
    def __init__(self, task_id: _Optional[str] = ..., signal_key: _Optional[str] = ...) -> None: ...

class DurableEvent(_message.Message):
    __slots__ = ("task_id", "signal_key", "data")
    TASK_ID_FIELD_NUMBER: _ClassVar[int]
    SIGNAL_KEY_FIELD_NUMBER: _ClassVar[int]
    DATA_FIELD_NUMBER: _ClassVar[int]
    task_id: str
    signal_key: str
    data: bytes
    def __init__(self, task_id: _Optional[str] = ..., signal_key: _Optional[str] = ..., data: _Optional[bytes] = ...) -> None: ...



import json
import traceback
from concurrent.futures import Future, ThreadPoolExecutor
from datetime import timedelta
from typing import TYPE_CHECKING, Any, cast
from warnings import warn

from hatchet_sdk.clients.admin import AdminClient
from hatchet_sdk.clients.dispatcher.dispatcher import (  # type: ignore[attr-defined]
    Action,
    DispatcherClient,
)
from hatchet_sdk.clients.events import EventClient
from hatchet_sdk.clients.listeners.durable_event_listener import (
    DurableEventListener,
    RegisterDurableEventRequest,
)
from hatchet_sdk.context.worker_context import WorkerContext
from hatchet_sdk.features.runs import RunsClient
from hatchet_sdk.logger import logger
from hatchet_sdk.utils.timedelta_to_expression import Duration, timedelta_to_expr
from hatchet_sdk.utils.typing import JSONSerializableMapping
from hatchet_sdk.waits import SleepCondition, UserEventCondition

if TYPE_CHECKING:
    from hatchet_sdk.runnables.task import Task
    from hatchet_sdk.runnables.types import R, TWorkflowInput


class Context:
    def __init__(
        self,
        action: Action,
        dispatcher_client: DispatcherClient,
        admin_client: AdminClient,
        event_client: EventClient,
        durable_event_listener: DurableEventListener | None,
        worker: WorkerContext,
        runs_client: RunsClient,
        lifespan_context: Any | None,
    ):
        self.worker = worker

        self.data = action.action_payload

        self.action = action

        self.step_run_id = action.step_run_id
        self.exit_flag = False
        self.dispatcher_client = dispatcher_client
        self.admin_client = admin_client
        self.event_client = event_client
        self.runs_client = runs_client
        self.durable_event_listener = durable_event_listener

        # FIXME: this limits the number of concurrent log requests to 1, which means we can do about
        # 100 log lines per second but this depends on network.
        self.logger_thread_pool = ThreadPoolExecutor(max_workers=1)
        self.stream_event_thread_pool = ThreadPoolExecutor(max_workers=1)

        self.input = self.data.input

        self._lifespan_context = lifespan_context

    def was_skipped(self, task: "Task[TWorkflowInput, R]") -> bool:
        return self.data.parents.get(task.name, {}).get("skipped", False)

    @property
    def trigger_data(self) -> JSONSerializableMapping:
        return self.data.triggers

    def _task_output(self, task: "Task[TWorkflowInput, R]") -> "R":
        from hatchet_sdk.runnables.types import R

        if self.was_skipped(task):
            raise ValueError("{task.name} was skipped")

        try:
            parent_step_data = cast(R, self.data.parents[task.name])
        except KeyError:
            raise ValueError(f"Step output for '{task.name}' not found")

        if parent_step_data and (v := task.validators.step_output):
            return cast(R, v.model_validate(parent_step_data))

        return parent_step_data

    def task_output(self, task: "Task[TWorkflowInput, R]") -> "R":
        from hatchet_sdk.runnables.types import R

        ## If the task is async, we need to wrap its output in a coroutine
        ## so that the type checker behaves right
        async def _aio_output() -> "R":
            return self._task_output(task)

        if task.is_async_function:
            return cast(R, _aio_output())

        return self._task_output(task)

    def aio_task_output(self, task: "Task[TWorkflowInput, R]") -> "R":
        warn(
            "`aio_task_output` is deprecated. Use `task_output` instead.",
            DeprecationWarning,
        )

        if task.is_async_function:
            return self.task_output(task)

        raise ValueError(
            f"Task '{task.name}' is not an async function. Use `task_output` instead."
        )

    @property
    def was_triggered_by_event(self) -> bool:
        return self.data.triggered_by == "event"

    @property
    def workflow_input(self) -> JSONSerializableMapping:
        return self.input

    @property
    def lifespan(self) -> Any:
        return self._lifespan_context

    @property
    def workflow_run_id(self) -> str:
        return self.action.workflow_run_id

    def cancel(self) -> None:
        logger.debug("cancelling step...")
        self.runs_client.cancel(self.step_run_id)
        self.exit_flag = True

    async def aio_cancel(self) -> None:
        logger.debug("cancelling step...")
        await self.runs_client.aio_cancel(self.step_run_id)
        self.exit_flag = True

    # done returns true if the context has been cancelled
    def done(self) -> bool:
        return self.exit_flag

    def _log(self, line: str) -> tuple[bool, Exception | None]:
        try:
            self.event_client.log(message=line, step_run_id=self.step_run_id)
            return True, None
        except Exception as e:
            # we don't want to raise an exception here, as it will kill the log thread
            return False, e

    def log(
        self, line: str | JSONSerializableMapping, raise_on_error: bool = False
    ) -> None:
        if self.step_run_id == "":
            return

        if not isinstance(line, str):
            try:
                line = json.dumps(line)
            except Exception:
                line = str(line)

        future = self.logger_thread_pool.submit(self._log, line)

        def handle_result(future: Future[tuple[bool, Exception | None]]) -> None:
            success, exception = future.result()

            if not success and exception:
                if raise_on_error:
                    raise exception
                else:
                    thread_trace = "".join(
                        traceback.format_exception(
                            type(exception), exception, exception.__traceback__
                        )
                    )
                    call_site_trace = "".join(traceback.format_stack())
                    logger.error(
                        f"Error in log thread: {exception}\n{thread_trace}\nCalled from:\n{call_site_trace}"
                    )

        future.add_done_callback(handle_result)

    def release_slot(self) -> None:
        return self.dispatcher_client.release_slot(self.step_run_id)

    def _put_stream(self, data: str | bytes) -> None:
        try:
            self.event_client.stream(data=data, step_run_id=self.step_run_id)
        except Exception as e:
            logger.error(f"Error putting stream event: {e}")

    def put_stream(self, data: str | bytes) -> None:
        if self.step_run_id == "":
            return

        self.stream_event_thread_pool.submit(self._put_stream, data)

    def refresh_timeout(self, increment_by: str | timedelta) -> None:
        if isinstance(increment_by, timedelta):
            increment_by = timedelta_to_expr(increment_by)

        try:
            return self.dispatcher_client.refresh_timeout(
                step_run_id=self.step_run_id, increment_by=increment_by
            )
        except Exception as e:
            logger.error(f"Error refreshing timeout: {e}")

    @property
    def retry_count(self) -> int:
        return self.action.retry_count

    @property
    def additional_metadata(self) -> JSONSerializableMapping | None:
        return self.action.additional_metadata

    @property
    def child_index(self) -> int | None:
        return self.action.child_workflow_index

    @property
    def child_key(self) -> str | None:
        return self.action.child_workflow_key

    @property
    def parent_workflow_run_id(self) -> str | None:
        return self.action.parent_workflow_run_id

    @property
    def task_run_errors(self) -> dict[str, str]:
        errors = self.data.step_run_errors

        if not errors:
            logger.error(
                "No step run errors found. `context.step_run_errors` is intended to be run in an on-failure step, and will only work on engine versions more recent than v0.53.10"
            )

        return errors

    def fetch_task_run_error(
        self,
        task: "Task[TWorkflowInput, R]",
    ) -> str | None:
        errors = self.data.step_run_errors

        return errors.get(task.name)


class DurableContext(Context):
    async def aio_wait_for(
        self, signal_key: str, *conditions: SleepCondition | UserEventCondition
    ) -> dict[str, Any]:
        if self.durable_event_listener is None:
            raise ValueError("Durable event listener is not available")

        task_id = self.step_run_id

        request = RegisterDurableEventRequest(
            task_id=task_id,
            signal_key=signal_key,
            conditions=list(conditions),
        )

        self.durable_event_listener.register_durable_event(request)

        return await self.durable_event_listener.result(
            task_id,
            signal_key,
        )

    async def aio_sleep_for(self, duration: Duration) -> dict[str, Any]:
        """
        Lightweight wrapper for durable sleep. Allows for shorthand usage of `ctx.aio_wait_for` when specifying a sleep condition.

        For more complicated conditions, use `ctx.aio_wait_for` directly.
        """

        return await self.aio_wait_for(
            f"sleep:{timedelta_to_expr(duration)}", SleepCondition(duration=duration)
        )



# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc
import warnings

from . import dispatcher_pb2 as dispatcher__pb2

GRPC_GENERATED_VERSION = '1.64.1'
GRPC_VERSION = grpc.__version__
EXPECTED_ERROR_RELEASE = '1.65.0'
SCHEDULED_RELEASE_DATE = 'June 25, 2024'
_version_not_supported = False

try:
    from grpc._utilities import first_version_is_lower
    _version_not_supported = first_version_is_lower(GRPC_VERSION, GRPC_GENERATED_VERSION)
except ImportError:
    _version_not_supported = True

if _version_not_supported:
    warnings.warn(
        f'The grpc package installed is at version {GRPC_VERSION},'
        + f' but the generated code in dispatcher_pb2_grpc.py depends on'
        + f' grpcio>={GRPC_GENERATED_VERSION}.'
        + f' Please upgrade your grpc module to grpcio>={GRPC_GENERATED_VERSION}'
        + f' or downgrade your generated code using grpcio-tools<={GRPC_VERSION}.'
        + f' This warning will become an error in {EXPECTED_ERROR_RELEASE},'
        + f' scheduled for release on {SCHEDULED_RELEASE_DATE}.',
        RuntimeWarning
    )


class DispatcherStub(object):
    """Missing associated documentation comment in .proto file."""

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.Register = channel.unary_unary(
                '/Dispatcher/Register',
                request_serializer=dispatcher__pb2.WorkerRegisterRequest.SerializeToString,
                response_deserializer=dispatcher__pb2.WorkerRegisterResponse.FromString,
                _registered_method=True)
        self.Listen = channel.unary_stream(
                '/Dispatcher/Listen',
                request_serializer=dispatcher__pb2.WorkerListenRequest.SerializeToString,
                response_deserializer=dispatcher__pb2.AssignedAction.FromString,
                _registered_method=True)
        self.ListenV2 = channel.unary_stream(
                '/Dispatcher/ListenV2',
                request_serializer=dispatcher__pb2.WorkerListenRequest.SerializeToString,
                response_deserializer=dispatcher__pb2.AssignedAction.FromString,
                _registered_method=True)
        self.Heartbeat = channel.unary_unary(
                '/Dispatcher/Heartbeat',
                request_serializer=dispatcher__pb2.HeartbeatRequest.SerializeToString,
                response_deserializer=dispatcher__pb2.HeartbeatResponse.FromString,
                _registered_method=True)
        self.SubscribeToWorkflowEvents = channel.unary_stream(
                '/Dispatcher/SubscribeToWorkflowEvents',
                request_serializer=dispatcher__pb2.SubscribeToWorkflowEventsRequest.SerializeToString,
                response_deserializer=dispatcher__pb2.WorkflowEvent.FromString,
                _registered_method=True)
        self.SubscribeToWorkflowRuns = channel.stream_stream(
                '/Dispatcher/SubscribeToWorkflowRuns',
                request_serializer=dispatcher__pb2.SubscribeToWorkflowRunsRequest.SerializeToString,
                response_deserializer=dispatcher__pb2.WorkflowRunEvent.FromString,
                _registered_method=True)
        self.SendStepActionEvent = channel.unary_unary(
                '/Dispatcher/SendStepActionEvent',
                request_serializer=dispatcher__pb2.StepActionEvent.SerializeToString,
                response_deserializer=dispatcher__pb2.ActionEventResponse.FromString,
                _registered_method=True)
        self.SendGroupKeyActionEvent = channel.unary_unary(
                '/Dispatcher/SendGroupKeyActionEvent',
                request_serializer=dispatcher__pb2.GroupKeyActionEvent.SerializeToString,
                response_deserializer=dispatcher__pb2.ActionEventResponse.FromString,
                _registered_method=True)
        self.PutOverridesData = channel.unary_unary(
                '/Dispatcher/PutOverridesData',
                request_serializer=dispatcher__pb2.OverridesData.SerializeToString,
                response_deserializer=dispatcher__pb2.OverridesDataResponse.FromString,
                _registered_method=True)
        self.Unsubscribe = channel.unary_unary(
                '/Dispatcher/Unsubscribe',
                request_serializer=dispatcher__pb2.WorkerUnsubscribeRequest.SerializeToString,
                response_deserializer=dispatcher__pb2.WorkerUnsubscribeResponse.FromString,
                _registered_method=True)
        self.RefreshTimeout = channel.unary_unary(
                '/Dispatcher/RefreshTimeout',
                request_serializer=dispatcher__pb2.RefreshTimeoutRequest.SerializeToString,
                response_deserializer=dispatcher__pb2.RefreshTimeoutResponse.FromString,
                _registered_method=True)
        self.ReleaseSlot = channel.unary_unary(
                '/Dispatcher/ReleaseSlot',
                request_serializer=dispatcher__pb2.ReleaseSlotRequest.SerializeToString,
                response_deserializer=dispatcher__pb2.ReleaseSlotResponse.FromString,
                _registered_method=True)
        self.UpsertWorkerLabels = channel.unary_unary(
                '/Dispatcher/UpsertWorkerLabels',
                request_serializer=dispatcher__pb2.UpsertWorkerLabelsRequest.SerializeToString,
                response_deserializer=dispatcher__pb2.UpsertWorkerLabelsResponse.FromString,
                _registered_method=True)


class DispatcherServicer(object):
    """Missing associated documentation comment in .proto file."""

    def Register(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def Listen(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def ListenV2(self, request, context):
        """ListenV2 is like listen, but implementation does not include heartbeats. This should only used by SDKs
        against engine version v0.18.1+
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def Heartbeat(self, request, context):
        """Heartbeat is a method for workers to send heartbeats to the dispatcher
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def SubscribeToWorkflowEvents(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def SubscribeToWorkflowRuns(self, request_iterator, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def SendStepActionEvent(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def SendGroupKeyActionEvent(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def PutOverridesData(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def Unsubscribe(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def RefreshTimeout(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def ReleaseSlot(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def UpsertWorkerLabels(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_DispatcherServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'Register': grpc.unary_unary_rpc_method_handler(
                    servicer.Register,
                    request_deserializer=dispatcher__pb2.WorkerRegisterRequest.FromString,
                    response_serializer=dispatcher__pb2.WorkerRegisterResponse.SerializeToString,
            ),
            'Listen': grpc.unary_stream_rpc_method_handler(
                    servicer.Listen,
                    request_deserializer=dispatcher__pb2.WorkerListenRequest.FromString,
                    response_serializer=dispatcher__pb2.AssignedAction.SerializeToString,
            ),
            'ListenV2': grpc.unary_stream_rpc_method_handler(
                    servicer.ListenV2,
                    request_deserializer=dispatcher__pb2.WorkerListenRequest.FromString,
                    response_serializer=dispatcher__pb2.AssignedAction.SerializeToString,
            ),
            'Heartbeat': grpc.unary_unary_rpc_method_handler(
                    servicer.Heartbeat,
                    request_deserializer=dispatcher__pb2.HeartbeatRequest.FromString,
                    response_serializer=dispatcher__pb2.HeartbeatResponse.SerializeToString,
            ),
            'SubscribeToWorkflowEvents': grpc.unary_stream_rpc_method_handler(
                    servicer.SubscribeToWorkflowEvents,
                    request_deserializer=dispatcher__pb2.SubscribeToWorkflowEventsRequest.FromString,
                    response_serializer=dispatcher__pb2.WorkflowEvent.SerializeToString,
            ),
            'SubscribeToWorkflowRuns': grpc.stream_stream_rpc_method_handler(
                    servicer.SubscribeToWorkflowRuns,
                    request_deserializer=dispatcher__pb2.SubscribeToWorkflowRunsRequest.FromString,
                    response_serializer=dispatcher__pb2.WorkflowRunEvent.SerializeToString,
            ),
            'SendStepActionEvent': grpc.unary_unary_rpc_method_handler(
                    servicer.SendStepActionEvent,
                    request_deserializer=dispatcher__pb2.StepActionEvent.FromString,
                    response_serializer=dispatcher__pb2.ActionEventResponse.SerializeToString,
            ),
            'SendGroupKeyActionEvent': grpc.unary_unary_rpc_method_handler(
                    servicer.SendGroupKeyActionEvent,
                    request_deserializer=dispatcher__pb2.GroupKeyActionEvent.FromString,
                    response_serializer=dispatcher__pb2.ActionEventResponse.SerializeToString,
            ),
            'PutOverridesData': grpc.unary_unary_rpc_method_handler(
                    servicer.PutOverridesData,
                    request_deserializer=dispatcher__pb2.OverridesData.FromString,
                    response_serializer=dispatcher__pb2.OverridesDataResponse.SerializeToString,
            ),
            'Unsubscribe': grpc.unary_unary_rpc_method_handler(
                    servicer.Unsubscribe,
                    request_deserializer=dispatcher__pb2.WorkerUnsubscribeRequest.FromString,
                    response_serializer=dispatcher__pb2.WorkerUnsubscribeResponse.SerializeToString,
            ),
            'RefreshTimeout': grpc.unary_unary_rpc_method_handler(
                    servicer.RefreshTimeout,
                    request_deserializer=dispatcher__pb2.RefreshTimeoutRequest.FromString,
                    response_serializer=dispatcher__pb2.RefreshTimeoutResponse.SerializeToString,
            ),
            'ReleaseSlot': grpc.unary_unary_rpc_method_handler(
                    servicer.ReleaseSlot,
                    request_deserializer=dispatcher__pb2.ReleaseSlotRequest.FromString,
                    response_serializer=dispatcher__pb2.ReleaseSlotResponse.SerializeToString,
            ),
            'UpsertWorkerLabels': grpc.unary_unary_rpc_method_handler(
                    servicer.UpsertWorkerLabels,
                    request_deserializer=dispatcher__pb2.UpsertWorkerLabelsRequest.FromString,
                    response_serializer=dispatcher__pb2.UpsertWorkerLabelsResponse.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'Dispatcher', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))
    server.add_registered_method_handlers('Dispatcher', rpc_method_handlers)


 # This class is part of an EXPERIMENTAL API.
class Dispatcher(object):
    """Missing associated documentation comment in .proto file."""

    @staticmethod
    def Register(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/Dispatcher/Register',
            dispatcher__pb2.WorkerRegisterRequest.SerializeToString,
            dispatcher__pb2.WorkerRegisterResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def Listen(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_stream(
            request,
            target,
            '/Dispatcher/Listen',
            dispatcher__pb2.WorkerListenRequest.SerializeToString,
            dispatcher__pb2.AssignedAction.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def ListenV2(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_stream(
            request,
            target,
            '/Dispatcher/ListenV2',
            dispatcher__pb2.WorkerListenRequest.SerializeToString,
            dispatcher__pb2.AssignedAction.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def Heartbeat(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/Dispatcher/Heartbeat',
            dispatcher__pb2.HeartbeatRequest.SerializeToString,
            dispatcher__pb2.HeartbeatResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def SubscribeToWorkflowEvents(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_stream(
            request,
            target,
            '/Dispatcher/SubscribeToWorkflowEvents',
            dispatcher__pb2.SubscribeToWorkflowEventsRequest.SerializeToString,
            dispatcher__pb2.WorkflowEvent.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def SubscribeToWorkflowRuns(request_iterator,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.stream_stream(
            request_iterator,
            target,
            '/Dispatcher/SubscribeToWorkflowRuns',
            dispatcher__pb2.SubscribeToWorkflowRunsRequest.SerializeToString,
            dispatcher__pb2.WorkflowRunEvent.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def SendStepActionEvent(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/Dispatcher/SendStepActionEvent',
            dispatcher__pb2.StepActionEvent.SerializeToString,
            dispatcher__pb2.ActionEventResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def SendGroupKeyActionEvent(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/Dispatcher/SendGroupKeyActionEvent',
            dispatcher__pb2.GroupKeyActionEvent.SerializeToString,
            dispatcher__pb2.ActionEventResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def PutOverridesData(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/Dispatcher/PutOverridesData',
            dispatcher__pb2.OverridesData.SerializeToString,
            dispatcher__pb2.OverridesDataResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def Unsubscribe(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/Dispatcher/Unsubscribe',
            dispatcher__pb2.WorkerUnsubscribeRequest.SerializeToString,
            dispatcher__pb2.WorkerUnsubscribeResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def RefreshTimeout(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/Dispatcher/RefreshTimeout',
            dispatcher__pb2.RefreshTimeoutRequest.SerializeToString,
            dispatcher__pb2.RefreshTimeoutResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def ReleaseSlot(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/Dispatcher/ReleaseSlot',
            dispatcher__pb2.ReleaseSlotRequest.SerializeToString,
            dispatcher__pb2.ReleaseSlotResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def UpsertWorkerLabels(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/Dispatcher/UpsertWorkerLabels',
            dispatcher__pb2.UpsertWorkerLabelsRequest.SerializeToString,
            dispatcher__pb2.UpsertWorkerLabelsResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)



# coding: utf-8

# flake8: noqa

"""
    Hatchet API

    The Hatchet API

    The version of the OpenAPI document: 1.0.0
    Generated by OpenAPI Generator (https://openapi-generator.tech)

    Do not edit the class manually.
"""  # noqa: E501


__version__ = "1.0.0"

# import apis into sdk package
from hatchet_sdk.v0.clients.rest.api.api_token_api import APITokenApi
from hatchet_sdk.v0.clients.rest.api.default_api import DefaultApi
from hatchet_sdk.v0.clients.rest.api.event_api import EventApi
from hatchet_sdk.v0.clients.rest.api.github_api import GithubApi
from hatchet_sdk.v0.clients.rest.api.healthcheck_api import HealthcheckApi
from hatchet_sdk.v0.clients.rest.api.log_api import LogApi
from hatchet_sdk.v0.clients.rest.api.metadata_api import MetadataApi
from hatchet_sdk.v0.clients.rest.api.rate_limits_api import RateLimitsApi
from hatchet_sdk.v0.clients.rest.api.slack_api import SlackApi
from hatchet_sdk.v0.clients.rest.api.sns_api import SNSApi
from hatchet_sdk.v0.clients.rest.api.step_run_api import StepRunApi
from hatchet_sdk.v0.clients.rest.api.tenant_api import TenantApi
from hatchet_sdk.v0.clients.rest.api.user_api import UserApi
from hatchet_sdk.v0.clients.rest.api.worker_api import WorkerApi
from hatchet_sdk.v0.clients.rest.api.workflow_api import WorkflowApi
from hatchet_sdk.v0.clients.rest.api.workflow_run_api import WorkflowRunApi
from hatchet_sdk.v0.clients.rest.api_client import ApiClient

# import ApiClient
from hatchet_sdk.v0.clients.rest.api_response import ApiResponse
from hatchet_sdk.v0.clients.rest.configuration import Configuration
from hatchet_sdk.v0.clients.rest.exceptions import (
    ApiAttributeError,
    ApiException,
    ApiKeyError,
    ApiTypeError,
    ApiValueError,
    OpenApiException,
)
from hatchet_sdk.v0.clients.rest.models.accept_invite_request import AcceptInviteRequest

# import models into sdk package
from hatchet_sdk.v0.clients.rest.models.api_error import APIError
from hatchet_sdk.v0.clients.rest.models.api_errors import APIErrors
from hatchet_sdk.v0.clients.rest.models.api_meta import APIMeta
from hatchet_sdk.v0.clients.rest.models.api_meta_auth import APIMetaAuth
from hatchet_sdk.v0.clients.rest.models.api_meta_integration import APIMetaIntegration
from hatchet_sdk.v0.clients.rest.models.api_meta_posthog import APIMetaPosthog
from hatchet_sdk.v0.clients.rest.models.api_resource_meta import APIResourceMeta
from hatchet_sdk.v0.clients.rest.models.api_token import APIToken
from hatchet_sdk.v0.clients.rest.models.bulk_create_event_request import (
    BulkCreateEventRequest,
)
from hatchet_sdk.v0.clients.rest.models.cancel_event_request import CancelEventRequest
from hatchet_sdk.v0.clients.rest.models.concurrency_limit_strategy import (
    ConcurrencyLimitStrategy,
)
from hatchet_sdk.v0.clients.rest.models.create_api_token_request import (
    CreateAPITokenRequest,
)
from hatchet_sdk.v0.clients.rest.models.create_api_token_response import (
    CreateAPITokenResponse,
)
from hatchet_sdk.v0.clients.rest.models.create_cron_workflow_trigger_request import (
    CreateCronWorkflowTriggerRequest,
)
from hatchet_sdk.v0.clients.rest.models.create_event_request import CreateEventRequest
from hatchet_sdk.v0.clients.rest.models.create_pull_request_from_step_run import (
    CreatePullRequestFromStepRun,
)
from hatchet_sdk.v0.clients.rest.models.create_sns_integration_request import (
    CreateSNSIntegrationRequest,
)
from hatchet_sdk.v0.clients.rest.models.create_tenant_alert_email_group_request import (
    CreateTenantAlertEmailGroupRequest,
)
from hatchet_sdk.v0.clients.rest.models.create_tenant_invite_request import (
    CreateTenantInviteRequest,
)
from hatchet_sdk.v0.clients.rest.models.create_tenant_request import CreateTenantRequest
from hatchet_sdk.v0.clients.rest.models.cron_workflows import CronWorkflows
from hatchet_sdk.v0.clients.rest.models.cron_workflows_list import CronWorkflowsList
from hatchet_sdk.v0.clients.rest.models.cron_workflows_method import CronWorkflowsMethod
from hatchet_sdk.v0.clients.rest.models.cron_workflows_order_by_field import (
    CronWorkflowsOrderByField,
)
from hatchet_sdk.v0.clients.rest.models.event import Event
from hatchet_sdk.v0.clients.rest.models.event_data import EventData
from hatchet_sdk.v0.clients.rest.models.event_key_list import EventKeyList
from hatchet_sdk.v0.clients.rest.models.event_list import EventList
from hatchet_sdk.v0.clients.rest.models.event_order_by_direction import (
    EventOrderByDirection,
)
from hatchet_sdk.v0.clients.rest.models.event_order_by_field import EventOrderByField
from hatchet_sdk.v0.clients.rest.models.event_update_cancel200_response import (
    EventUpdateCancel200Response,
)
from hatchet_sdk.v0.clients.rest.models.event_workflow_run_summary import (
    EventWorkflowRunSummary,
)
from hatchet_sdk.v0.clients.rest.models.events import Events
from hatchet_sdk.v0.clients.rest.models.get_step_run_diff_response import (
    GetStepRunDiffResponse,
)
from hatchet_sdk.v0.clients.rest.models.info_get_version200_response import (
    InfoGetVersion200Response,
)
from hatchet_sdk.v0.clients.rest.models.job import Job
from hatchet_sdk.v0.clients.rest.models.job_run import JobRun
from hatchet_sdk.v0.clients.rest.models.job_run_status import JobRunStatus
from hatchet_sdk.v0.clients.rest.models.list_api_tokens_response import (
    ListAPITokensResponse,
)
from hatchet_sdk.v0.clients.rest.models.list_pull_requests_response import (
    ListPullRequestsResponse,
)
from hatchet_sdk.v0.clients.rest.models.list_slack_webhooks import ListSlackWebhooks
from hatchet_sdk.v0.clients.rest.models.list_sns_integrations import ListSNSIntegrations
from hatchet_sdk.v0.clients.rest.models.log_line import LogLine
from hatchet_sdk.v0.clients.rest.models.log_line_level import LogLineLevel
from hatchet_sdk.v0.clients.rest.models.log_line_list import LogLineList
from hatchet_sdk.v0.clients.rest.models.log_line_order_by_direction import (
    LogLineOrderByDirection,
)
from hatchet_sdk.v0.clients.rest.models.log_line_order_by_field import (
    LogLineOrderByField,
)
from hatchet_sdk.v0.clients.rest.models.pagination_response import PaginationResponse
from hatchet_sdk.v0.clients.rest.models.pull_request import PullRequest
from hatchet_sdk.v0.clients.rest.models.pull_request_state import PullRequestState
from hatchet_sdk.v0.clients.rest.models.queue_metrics import QueueMetrics
from hatchet_sdk.v0.clients.rest.models.rate_limit import RateLimit
from hatchet_sdk.v0.clients.rest.models.rate_limit_list import RateLimitList
from hatchet_sdk.v0.clients.rest.models.rate_limit_order_by_direction import (
    RateLimitOrderByDirection,
)
from hatchet_sdk.v0.clients.rest.models.rate_limit_order_by_field import (
    RateLimitOrderByField,
)
from hatchet_sdk.v0.clients.rest.models.recent_step_runs import RecentStepRuns
from hatchet_sdk.v0.clients.rest.models.reject_invite_request import RejectInviteRequest
from hatchet_sdk.v0.clients.rest.models.replay_event_request import ReplayEventRequest
from hatchet_sdk.v0.clients.rest.models.replay_workflow_runs_request import (
    ReplayWorkflowRunsRequest,
)
from hatchet_sdk.v0.clients.rest.models.replay_workflow_runs_response import (
    ReplayWorkflowRunsResponse,
)
from hatchet_sdk.v0.clients.rest.models.rerun_step_run_request import (
    RerunStepRunRequest,
)
from hatchet_sdk.v0.clients.rest.models.schedule_workflow_run_request import (
    ScheduleWorkflowRunRequest,
)
from hatchet_sdk.v0.clients.rest.models.scheduled_run_status import ScheduledRunStatus
from hatchet_sdk.v0.clients.rest.models.scheduled_workflows import ScheduledWorkflows
from hatchet_sdk.v0.clients.rest.models.scheduled_workflows_list import (
    ScheduledWorkflowsList,
)
from hatchet_sdk.v0.clients.rest.models.scheduled_workflows_method import (
    ScheduledWorkflowsMethod,
)
from hatchet_sdk.v0.clients.rest.models.scheduled_workflows_order_by_field import (
    ScheduledWorkflowsOrderByField,
)
from hatchet_sdk.v0.clients.rest.models.semaphore_slots import SemaphoreSlots
from hatchet_sdk.v0.clients.rest.models.slack_webhook import SlackWebhook
from hatchet_sdk.v0.clients.rest.models.sns_integration import SNSIntegration
from hatchet_sdk.v0.clients.rest.models.step import Step
from hatchet_sdk.v0.clients.rest.models.step_run import StepRun
from hatchet_sdk.v0.clients.rest.models.step_run_archive import StepRunArchive
from hatchet_sdk.v0.clients.rest.models.step_run_archive_list import StepRunArchiveList
from hatchet_sdk.v0.clients.rest.models.step_run_diff import StepRunDiff
from hatchet_sdk.v0.clients.rest.models.step_run_event import StepRunEvent
from hatchet_sdk.v0.clients.rest.models.step_run_event_list import StepRunEventList
from hatchet_sdk.v0.clients.rest.models.step_run_event_reason import StepRunEventReason
from hatchet_sdk.v0.clients.rest.models.step_run_event_severity import (
    StepRunEventSeverity,
)
from hatchet_sdk.v0.clients.rest.models.step_run_status import StepRunStatus
from hatchet_sdk.v0.clients.rest.models.tenant import Tenant
from hatchet_sdk.v0.clients.rest.models.tenant_alert_email_group import (
    TenantAlertEmailGroup,
)
from hatchet_sdk.v0.clients.rest.models.tenant_alert_email_group_list import (
    TenantAlertEmailGroupList,
)
from hatchet_sdk.v0.clients.rest.models.tenant_alerting_settings import (
    TenantAlertingSettings,
)
from hatchet_sdk.v0.clients.rest.models.tenant_invite import TenantInvite
from hatchet_sdk.v0.clients.rest.models.tenant_invite_list import TenantInviteList
from hatchet_sdk.v0.clients.rest.models.tenant_list import TenantList
from hatchet_sdk.v0.clients.rest.models.tenant_member import TenantMember
from hatchet_sdk.v0.clients.rest.models.tenant_member_list import TenantMemberList
from hatchet_sdk.v0.clients.rest.models.tenant_member_role import TenantMemberRole
from hatchet_sdk.v0.clients.rest.models.tenant_queue_metrics import TenantQueueMetrics
from hatchet_sdk.v0.clients.rest.models.tenant_resource import TenantResource
from hatchet_sdk.v0.clients.rest.models.tenant_resource_limit import TenantResourceLimit
from hatchet_sdk.v0.clients.rest.models.tenant_resource_policy import (
    TenantResourcePolicy,
)
from hatchet_sdk.v0.clients.rest.models.tenant_step_run_queue_metrics import (
    TenantStepRunQueueMetrics,
)
from hatchet_sdk.v0.clients.rest.models.trigger_workflow_run_request import (
    TriggerWorkflowRunRequest,
)
from hatchet_sdk.v0.clients.rest.models.update_tenant_alert_email_group_request import (
    UpdateTenantAlertEmailGroupRequest,
)
from hatchet_sdk.v0.clients.rest.models.update_tenant_invite_request import (
    UpdateTenantInviteRequest,
)
from hatchet_sdk.v0.clients.rest.models.update_tenant_request import UpdateTenantRequest
from hatchet_sdk.v0.clients.rest.models.update_worker_request import UpdateWorkerRequest
from hatchet_sdk.v0.clients.rest.models.user import User
from hatchet_sdk.v0.clients.rest.models.user_change_password_request import (
    UserChangePasswordRequest,
)
from hatchet_sdk.v0.clients.rest.models.user_login_request import UserLoginRequest
from hatchet_sdk.v0.clients.rest.models.user_register_request import UserRegisterRequest
from hatchet_sdk.v0.clients.rest.models.user_tenant_memberships_list import (
    UserTenantMembershipsList,
)
from hatchet_sdk.v0.clients.rest.models.user_tenant_public import UserTenantPublic
from hatchet_sdk.v0.clients.rest.models.webhook_worker import WebhookWorker
from hatchet_sdk.v0.clients.rest.models.webhook_worker_create_request import (
    WebhookWorkerCreateRequest,
)
from hatchet_sdk.v0.clients.rest.models.webhook_worker_create_response import (
    WebhookWorkerCreateResponse,
)
from hatchet_sdk.v0.clients.rest.models.webhook_worker_created import (
    WebhookWorkerCreated,
)
from hatchet_sdk.v0.clients.rest.models.webhook_worker_list_response import (
    WebhookWorkerListResponse,
)
from hatchet_sdk.v0.clients.rest.models.webhook_worker_request import (
    WebhookWorkerRequest,
)
from hatchet_sdk.v0.clients.rest.models.webhook_worker_request_list_response import (
    WebhookWorkerRequestListResponse,
)
from hatchet_sdk.v0.clients.rest.models.webhook_worker_request_method import (
    WebhookWorkerRequestMethod,
)
from hatchet_sdk.v0.clients.rest.models.worker import Worker
from hatchet_sdk.v0.clients.rest.models.worker_label import WorkerLabel
from hatchet_sdk.v0.clients.rest.models.worker_list import WorkerList
from hatchet_sdk.v0.clients.rest.models.worker_runtime_info import WorkerRuntimeInfo
from hatchet_sdk.v0.clients.rest.models.worker_runtime_sdks import WorkerRuntimeSDKs
from hatchet_sdk.v0.clients.rest.models.worker_type import WorkerType
from hatchet_sdk.v0.clients.rest.models.workflow import Workflow
from hatchet_sdk.v0.clients.rest.models.workflow_concurrency import WorkflowConcurrency
from hatchet_sdk.v0.clients.rest.models.workflow_kind import WorkflowKind
from hatchet_sdk.v0.clients.rest.models.workflow_list import WorkflowList
from hatchet_sdk.v0.clients.rest.models.workflow_metrics import WorkflowMetrics
from hatchet_sdk.v0.clients.rest.models.workflow_run import WorkflowRun
from hatchet_sdk.v0.clients.rest.models.workflow_run_list import WorkflowRunList
from hatchet_sdk.v0.clients.rest.models.workflow_run_order_by_direction import (
    WorkflowRunOrderByDirection,
)
from hatchet_sdk.v0.clients.rest.models.workflow_run_order_by_field import (
    WorkflowRunOrderByField,
)
from hatchet_sdk.v0.clients.rest.models.workflow_run_shape import WorkflowRunShape
from hatchet_sdk.v0.clients.rest.models.workflow_run_status import WorkflowRunStatus
from hatchet_sdk.v0.clients.rest.models.workflow_run_triggered_by import (
    WorkflowRunTriggeredBy,
)
from hatchet_sdk.v0.clients.rest.models.workflow_runs_cancel_request import (
    WorkflowRunsCancelRequest,
)
from hatchet_sdk.v0.clients.rest.models.workflow_runs_metrics import WorkflowRunsMetrics
from hatchet_sdk.v0.clients.rest.models.workflow_runs_metrics_counts import (
    WorkflowRunsMetricsCounts,
)
from hatchet_sdk.v0.clients.rest.models.workflow_tag import WorkflowTag
from hatchet_sdk.v0.clients.rest.models.workflow_trigger_cron_ref import (
    WorkflowTriggerCronRef,
)
from hatchet_sdk.v0.clients.rest.models.workflow_trigger_event_ref import (
    WorkflowTriggerEventRef,
)
from hatchet_sdk.v0.clients.rest.models.workflow_triggers import WorkflowTriggers
from hatchet_sdk.v0.clients.rest.models.workflow_update_request import (
    WorkflowUpdateRequest,
)
from hatchet_sdk.v0.clients.rest.models.workflow_version import WorkflowVersion
from hatchet_sdk.v0.clients.rest.models.workflow_version_definition import (
    WorkflowVersionDefinition,
)
from hatchet_sdk.v0.clients.rest.models.workflow_version_meta import WorkflowVersionMeta
from hatchet_sdk.v0.clients.rest.models.workflow_workers_count import (
    WorkflowWorkersCount,
)



from hatchet_sdk.v0.clients.rest.models.accept_invite_request import AcceptInviteRequest

# import models into sdk package
from hatchet_sdk.v0.clients.rest.models.api_error import APIError
from hatchet_sdk.v0.clients.rest.models.api_errors import APIErrors
from hatchet_sdk.v0.clients.rest.models.api_meta import APIMeta
from hatchet_sdk.v0.clients.rest.models.api_meta_auth import APIMetaAuth
from hatchet_sdk.v0.clients.rest.models.api_meta_integration import APIMetaIntegration
from hatchet_sdk.v0.clients.rest.models.api_resource_meta import APIResourceMeta
from hatchet_sdk.v0.clients.rest.models.api_token import APIToken
from hatchet_sdk.v0.clients.rest.models.create_api_token_request import (
    CreateAPITokenRequest,
)
from hatchet_sdk.v0.clients.rest.models.create_api_token_response import (
    CreateAPITokenResponse,
)
from hatchet_sdk.v0.clients.rest.models.create_pull_request_from_step_run import (
    CreatePullRequestFromStepRun,
)
from hatchet_sdk.v0.clients.rest.models.create_tenant_invite_request import (
    CreateTenantInviteRequest,
)
from hatchet_sdk.v0.clients.rest.models.create_tenant_request import CreateTenantRequest
from hatchet_sdk.v0.clients.rest.models.event import Event
from hatchet_sdk.v0.clients.rest.models.event_data import EventData
from hatchet_sdk.v0.clients.rest.models.event_key_list import EventKeyList
from hatchet_sdk.v0.clients.rest.models.event_list import EventList
from hatchet_sdk.v0.clients.rest.models.event_order_by_direction import (
    EventOrderByDirection,
)
from hatchet_sdk.v0.clients.rest.models.event_order_by_field import EventOrderByField
from hatchet_sdk.v0.clients.rest.models.event_workflow_run_summary import (
    EventWorkflowRunSummary,
)
from hatchet_sdk.v0.clients.rest.models.get_step_run_diff_response import (
    GetStepRunDiffResponse,
)
from hatchet_sdk.v0.clients.rest.models.github_app_installation import (
    GithubAppInstallation,
)
from hatchet_sdk.v0.clients.rest.models.github_branch import GithubBranch
from hatchet_sdk.v0.clients.rest.models.github_repo import GithubRepo
from hatchet_sdk.v0.clients.rest.models.job import Job
from hatchet_sdk.v0.clients.rest.models.job_run import JobRun
from hatchet_sdk.v0.clients.rest.models.job_run_status import JobRunStatus
from hatchet_sdk.v0.clients.rest.models.link_github_repository_request import (
    LinkGithubRepositoryRequest,
)
from hatchet_sdk.v0.clients.rest.models.list_api_tokens_response import (
    ListAPITokensResponse,
)
from hatchet_sdk.v0.clients.rest.models.list_github_app_installations_response import (
    ListGithubAppInstallationsResponse,
)
from hatchet_sdk.v0.clients.rest.models.list_pull_requests_response import (
    ListPullRequestsResponse,
)
from hatchet_sdk.v0.clients.rest.models.log_line import LogLine
from hatchet_sdk.v0.clients.rest.models.log_line_level import LogLineLevel
from hatchet_sdk.v0.clients.rest.models.log_line_list import LogLineList
from hatchet_sdk.v0.clients.rest.models.log_line_order_by_direction import (
    LogLineOrderByDirection,
)
from hatchet_sdk.v0.clients.rest.models.log_line_order_by_field import (
    LogLineOrderByField,
)
from hatchet_sdk.v0.clients.rest.models.pagination_response import PaginationResponse
from hatchet_sdk.v0.clients.rest.models.pull_request import PullRequest
from hatchet_sdk.v0.clients.rest.models.pull_request_state import PullRequestState
from hatchet_sdk.v0.clients.rest.models.reject_invite_request import RejectInviteRequest
from hatchet_sdk.v0.clients.rest.models.replay_event_request import ReplayEventRequest
from hatchet_sdk.v0.clients.rest.models.rerun_step_run_request import (
    RerunStepRunRequest,
)
from hatchet_sdk.v0.clients.rest.models.step import Step
from hatchet_sdk.v0.clients.rest.models.step_run import StepRun
from hatchet_sdk.v0.clients.rest.models.step_run_diff import StepRunDiff
from hatchet_sdk.v0.clients.rest.models.step_run_status import StepRunStatus
from hatchet_sdk.v0.clients.rest.models.tenant import Tenant
from hatchet_sdk.v0.clients.rest.models.tenant_invite import TenantInvite
from hatchet_sdk.v0.clients.rest.models.tenant_invite_list import TenantInviteList
from hatchet_sdk.v0.clients.rest.models.tenant_list import TenantList
from hatchet_sdk.v0.clients.rest.models.tenant_member import TenantMember
from hatchet_sdk.v0.clients.rest.models.tenant_member_list import TenantMemberList
from hatchet_sdk.v0.clients.rest.models.tenant_member_role import TenantMemberRole
from hatchet_sdk.v0.clients.rest.models.trigger_workflow_run_request import (
    TriggerWorkflowRunRequest,
)
from hatchet_sdk.v0.clients.rest.models.update_tenant_invite_request import (
    UpdateTenantInviteRequest,
)
from hatchet_sdk.v0.clients.rest.models.user import User
from hatchet_sdk.v0.clients.rest.models.user_login_request import UserLoginRequest
from hatchet_sdk.v0.clients.rest.models.user_register_request import UserRegisterRequest
from hatchet_sdk.v0.clients.rest.models.user_tenant_memberships_list import (
    UserTenantMembershipsList,
)
from hatchet_sdk.v0.clients.rest.models.user_tenant_public import UserTenantPublic
from hatchet_sdk.v0.clients.rest.models.worker_list import WorkerList
from hatchet_sdk.v0.clients.rest.models.workflow import Workflow
from hatchet_sdk.v0.clients.rest.models.workflow_deployment_config import (
    WorkflowDeploymentConfig,
)
from hatchet_sdk.v0.clients.rest.models.workflow_list import WorkflowList
from hatchet_sdk.v0.clients.rest.models.workflow_run import WorkflowRun
from hatchet_sdk.v0.clients.rest.models.workflow_run_list import WorkflowRunList
from hatchet_sdk.v0.clients.rest.models.workflow_run_status import WorkflowRunStatus
from hatchet_sdk.v0.clients.rest.models.workflow_run_triggered_by import (
    WorkflowRunTriggeredBy,
)
from hatchet_sdk.v0.clients.rest.models.workflow_tag import WorkflowTag
from hatchet_sdk.v0.clients.rest.models.workflow_trigger_cron_ref import (
    WorkflowTriggerCronRef,
)
from hatchet_sdk.v0.clients.rest.models.workflow_trigger_event_ref import (
    WorkflowTriggerEventRef,
)
from hatchet_sdk.v0.clients.rest.models.workflow_triggers import WorkflowTriggers
from hatchet_sdk.v0.clients.rest.models.workflow_version import WorkflowVersion
from hatchet_sdk.v0.clients.rest.models.workflow_version_definition import (
    WorkflowVersionDefinition,
)
from hatchet_sdk.v0.clients.rest.models.workflow_version_meta import WorkflowVersionMeta
from hatchet_sdk.v0.contracts.workflows_pb2 import (
    ConcurrencyLimitStrategy,
    CreateWorkflowVersionOpts,
    RateLimitDuration,
    StickyStrategy,
    WorkerLabelComparator,
)
from hatchet_sdk.v0.utils.aio_utils import sync_to_async

from .client import new_client
from .clients.admin import (
    ChildTriggerWorkflowOptions,
    DedupeViolationErr,
    ScheduleTriggerWorkflowOptions,
    TriggerWorkflowOptions,
)
from .clients.events import PushEventOptions
from .clients.run_event_listener import StepRunEventType, WorkflowRunEventType
from .context.context import Context
from .context.worker_context import WorkerContext
from .hatchet import ClientConfig, Hatchet, concurrency, on_failure_step, step, workflow
from .worker import Worker, WorkerStartOptions, WorkerStatus
from .workflow import ConcurrencyExpression

__all__ = [
    "AcceptInviteRequest",
    "APIError",
    "APIErrors",
    "APIMeta",
    "APIMetaAuth",
    "APIMetaIntegration",
    "APIResourceMeta",
    "APIToken",
    "CreateAPITokenRequest",
    "CreateAPITokenResponse",
    "CreatePullRequestFromStepRun",
    "CreateTenantInviteRequest",
    "CreateTenantRequest",
    "Event",
    "EventData",
    "EventKeyList",
    "EventList",
    "EventOrderByDirection",
    "EventOrderByField",
    "EventWorkflowRunSummary",
    "GetStepRunDiffResponse",
    "GithubAppInstallation",
    "GithubBranch",
    "GithubRepo",
    "Job",
    "JobRun",
    "JobRunStatus",
    "LinkGithubRepositoryRequest",
    "ListAPITokensResponse",
    "ListGithubAppInstallationsResponse",
    "ListPullRequestsResponse",
    "LogLine",
    "LogLineLevel",
    "LogLineList",
    "LogLineOrderByDirection",
    "LogLineOrderByField",
    "PaginationResponse",
    "PullRequest",
    "PullRequestState",
    "RejectInviteRequest",
    "ReplayEventRequest",
    "RerunStepRunRequest",
    "Step",
    "StepRun",
    "StepRunDiff",
    "StepRunStatus",
    "sync_to_async",
    "Tenant",
    "TenantInvite",
    "TenantInviteList",
    "TenantList",
    "TenantMember",
    "TenantMemberList",
    "TenantMemberRole",
    "TriggerWorkflowRunRequest",
    "UpdateTenantInviteRequest",
    "User",
    "UserLoginRequest",
    "UserRegisterRequest",
    "UserTenantMembershipsList",
    "UserTenantPublic",
    "Worker",
    "WorkerLabelComparator",
    "WorkerList",
    "Workflow",
    "WorkflowDeploymentConfig",
    "WorkflowList",
    "WorkflowRun",
    "WorkflowRunList",
    "WorkflowRunStatus",
    "WorkflowRunTriggeredBy",
    "WorkflowTag",
    "WorkflowTriggerCronRef",
    "WorkflowTriggerEventRef",
    "WorkflowTriggers",
    "WorkflowVersion",
    "WorkflowVersionDefinition",
    "WorkflowVersionMeta",
    "ConcurrencyLimitStrategy",
    "CreateWorkflowVersionOpts",
    "RateLimitDuration",
    "StickyStrategy",
    "new_client",
    "ChildTriggerWorkflowOptions",
    "DedupeViolationErr",
    "ScheduleTriggerWorkflowOptions",
    "TriggerWorkflowOptions",
    "PushEventOptions",
    "StepRunEventType",
    "WorkflowRunEventType",
    "Context",
    "WorkerContext",
    "ClientConfig",
    "Hatchet",
    "concurrency",
    "on_failure_step",
    "step",
    "workflow",
    "Worker",
    "WorkerStartOptions",
    "WorkerStatus",
    "ConcurrencyExpression",
]



import functools
import logging
from concurrent.futures import ThreadPoolExecutor
from contextvars import ContextVar
from io import StringIO
from typing import Any, Awaitable, Callable, ItemsView, ParamSpec, TypeVar

from hatchet_sdk.clients.events import EventClient
from hatchet_sdk.logger import logger
from hatchet_sdk.runnables.contextvars import ctx_step_run_id, ctx_workflow_run_id

T = TypeVar("T")
P = ParamSpec("P")


def copy_context_vars(
    ctx_vars: ItemsView[ContextVar[Any], Any],
    func: Callable[P, T],
    *args: P.args,
    **kwargs: P.kwargs,
) -> T:
    for var, value in ctx_vars:
        var.set(value)
    return func(*args, **kwargs)


class InjectingFilter(logging.Filter):
    # For some reason, only the InjectingFilter has access to the contextvars method sr.get(),
    # otherwise we would use emit within the CustomLogHandler
    def filter(self, record: logging.LogRecord) -> bool:
        ## TODO: Change how we do this to not assign to the log record
        record.workflow_run_id = ctx_workflow_run_id.get()
        record.step_run_id = ctx_step_run_id.get()
        return True


class CustomLogHandler(logging.StreamHandler):  # type: ignore[type-arg]
    def __init__(self, event_client: EventClient, stream: StringIO | None = None):
        super().__init__(stream)
        self.logger_thread_pool = ThreadPoolExecutor(max_workers=1)
        self.event_client = event_client

    def _log(self, line: str, step_run_id: str | None) -> None:
        try:
            if not step_run_id:
                return

            self.event_client.log(message=line, step_run_id=step_run_id)
        except Exception as e:
            logger.error(f"Error logging: {e}")

    def emit(self, record: logging.LogRecord) -> None:
        super().emit(record)

        log_entry = self.format(record)

        ## TODO: Change how we do this to not assign to the log record
        self.logger_thread_pool.submit(self._log, log_entry, record.step_run_id)  # type: ignore


def capture_logs(
    logger: logging.Logger, event_client: "EventClient", func: Callable[P, Awaitable[T]]
) -> Callable[P, Awaitable[T]]:
    @functools.wraps(func)
    async def wrapper(*args: P.args, **kwargs: P.kwargs) -> T:
        log_stream = StringIO()
        custom_handler = CustomLogHandler(event_client, log_stream)
        custom_handler.setLevel(logging.INFO)
        custom_handler.addFilter(InjectingFilter())
        logger.addHandler(custom_handler)

        try:
            result = await func(*args, **kwargs)
        finally:
            custom_handler.flush()
            logger.removeHandler(custom_handler)
            log_stream.close()

        return result

    return wrapper



from hatchet_sdk.clients.dispatcher.dispatcher import DispatcherClient


class WorkerContext:
    _worker_id: str | None = None
    _registered_workflow_names: list[str] = []
    _labels: dict[str, str | int] = {}

    def __init__(self, labels: dict[str, str | int], client: DispatcherClient):
        self._labels = labels
        self.client = client

    def labels(self) -> dict[str, str | int]:
        return self._labels

    def upsert_labels(self, labels: dict[str, str | int]) -> None:
        self.client.upsert_worker_labels(self._worker_id, labels)
        self._labels.update(labels)

    async def async_upsert_labels(self, labels: dict[str, str | int]) -> None:
        await self.client.async_upsert_worker_labels(self._worker_id, labels)
        self._labels.update(labels)

    def id(self) -> str | None:
        return self._worker_id

    # def has_workflow(self, workflow_name: str):
    #     return workflow_name in self._registered_workflow_names



import asyncio
import datetime
import json
from typing import Any, Dict, List, Optional, TypedDict
from uuid import uuid4

import grpc
from google.protobuf import timestamp_pb2

from hatchet_sdk.v0.clients.rest.tenacity_utils import tenacity_retry
from hatchet_sdk.v0.contracts.events_pb2 import (
    BulkPushEventRequest,
    Event,
    PushEventRequest,
    PutLogRequest,
    PutStreamEventRequest,
)
from hatchet_sdk.v0.contracts.events_pb2_grpc import EventsServiceStub

from ..loader import ClientConfig
from ..metadata import get_metadata


def new_event(conn, config: ClientConfig):
    return EventClient(
        client=EventsServiceStub(conn),
        config=config,
    )


def proto_timestamp_now():
    t = datetime.datetime.now().timestamp()
    seconds = int(t)
    nanos = int(t % 1 * 1e9)

    return timestamp_pb2.Timestamp(seconds=seconds, nanos=nanos)


class PushEventOptions(TypedDict, total=False):
    additional_metadata: Dict[str, str] | None = None
    namespace: str | None = None


class BulkPushEventOptions(TypedDict, total=False):
    namespace: str | None = None


class BulkPushEventWithMetadata(TypedDict, total=False):
    key: str
    payload: Any
    additional_metadata: Optional[Dict[str, Any]]  # Optional metadata


class EventClient:
    def __init__(self, client: EventsServiceStub, config: ClientConfig):
        self.client = client
        self.token = config.token
        self.namespace = config.namespace

    async def async_push(
        self, event_key, payload, options: Optional[PushEventOptions] = None
    ) -> Event:
        return await asyncio.to_thread(
            self.push, event_key=event_key, payload=payload, options=options
        )

    async def async_bulk_push(
        self,
        events: List[BulkPushEventWithMetadata],
        options: Optional[BulkPushEventOptions] = None,
    ) -> List[Event]:
        return await asyncio.to_thread(self.bulk_push, events=events, options=options)

    ## IMPORTANT: Keep this method's signature in sync with the wrapper in the OTel instrumentor
    @tenacity_retry
    def push(self, event_key, payload, options: PushEventOptions = None) -> Event:
        namespace = self.namespace

        if (
            options is not None
            and "namespace" in options
            and options["namespace"] is not None
        ):
            namespace = options.pop("namespace")

        namespaced_event_key = namespace + event_key

        try:
            meta = dict() if options is None else options["additional_metadata"]
            meta_bytes = None if meta is None else json.dumps(meta).encode("utf-8")
        except Exception as e:
            raise ValueError(f"Error encoding meta: {e}")

        try:
            payload_bytes = json.dumps(payload).encode("utf-8")
        except json.UnicodeEncodeError as e:
            raise ValueError(f"Error encoding payload: {e}")

        request = PushEventRequest(
            key=namespaced_event_key,
            payload=payload_bytes,
            eventTimestamp=proto_timestamp_now(),
            additionalMetadata=meta_bytes,
        )

        return self.client.Push(request, metadata=get_metadata(self.token))

    ## IMPORTANT: Keep this method's signature in sync with the wrapper in the OTel instrumentor
    @tenacity_retry
    def bulk_push(
        self,
        events: List[BulkPushEventWithMetadata],
        options: BulkPushEventOptions = None,
    ) -> List[Event]:
        namespace = self.namespace

        if (
            options is not None
            and "namespace" in options
            and options["namespace"] is not None
        ):
            namespace = options.pop("namespace")

        bulk_events = []
        for event in events:
            event_key = namespace + event["key"]
            payload = event["payload"]

            try:
                meta = event.get("additional_metadata", {})
                meta_bytes = json.dumps(meta).encode("utf-8") if meta else None
            except Exception as e:
                raise ValueError(f"Error encoding meta: {e}")

            try:
                payload_bytes = json.dumps(payload).encode("utf-8")
            except json.UnicodeEncodeError as e:
                raise ValueError(f"Error encoding payload: {e}")

            request = PushEventRequest(
                key=event_key,
                payload=payload_bytes,
                eventTimestamp=proto_timestamp_now(),
                additionalMetadata=meta_bytes,
            )
            bulk_events.append(request)

        bulk_request = BulkPushEventRequest(events=bulk_events)

        response = self.client.BulkPush(bulk_request, metadata=get_metadata(self.token))

        return response.events

    def log(self, message: str, step_run_id: str):
        try:
            request = PutLogRequest(
                stepRunId=step_run_id,
                createdAt=proto_timestamp_now(),
                message=message,
            )

            self.client.PutLog(request, metadata=get_metadata(self.token))
        except Exception as e:
            raise ValueError(f"Error logging: {e}")

    def stream(self, data: str | bytes, step_run_id: str):
        try:
            if isinstance(data, str):
                data_bytes = data.encode("utf-8")
            elif isinstance(data, bytes):
                data_bytes = data
            else:
                raise ValueError("Invalid data type. Expected str, bytes, or file.")

            request = PutStreamEventRequest(
                stepRunId=step_run_id,
                createdAt=proto_timestamp_now(),
                message=data_bytes,
            )
            self.client.PutStreamEvent(request, metadata=get_metadata(self.token))
        except Exception as e:
            raise ValueError(f"Error putting stream event: {e}")



import asyncio
from enum import Enum
from queue import Empty, Queue
from threading import Thread
from typing import Any, AsyncGenerator, Callable, Generator, Literal, TypeVar, cast

import grpc
from pydantic import BaseModel

from hatchet_sdk.config import ClientConfig
from hatchet_sdk.connection import new_conn
from hatchet_sdk.contracts.dispatcher_pb2 import (
    RESOURCE_TYPE_STEP_RUN,
    RESOURCE_TYPE_WORKFLOW_RUN,
    ResourceEventType,
    SubscribeToWorkflowEventsRequest,
    WorkflowEvent,
)
from hatchet_sdk.contracts.dispatcher_pb2_grpc import DispatcherStub
from hatchet_sdk.metadata import get_metadata

DEFAULT_ACTION_LISTENER_RETRY_INTERVAL = 5  # seconds
DEFAULT_ACTION_LISTENER_RETRY_COUNT = 5


class StepRunEventType(str, Enum):
    STEP_RUN_EVENT_TYPE_STARTED = "STEP_RUN_EVENT_TYPE_STARTED"
    STEP_RUN_EVENT_TYPE_COMPLETED = "STEP_RUN_EVENT_TYPE_COMPLETED"
    STEP_RUN_EVENT_TYPE_FAILED = "STEP_RUN_EVENT_TYPE_FAILED"
    STEP_RUN_EVENT_TYPE_CANCELLED = "STEP_RUN_EVENT_TYPE_CANCELLED"
    STEP_RUN_EVENT_TYPE_TIMED_OUT = "STEP_RUN_EVENT_TYPE_TIMED_OUT"
    STEP_RUN_EVENT_TYPE_STREAM = "STEP_RUN_EVENT_TYPE_STREAM"


class WorkflowRunEventType(str, Enum):
    WORKFLOW_RUN_EVENT_TYPE_STARTED = "WORKFLOW_RUN_EVENT_TYPE_STARTED"
    WORKFLOW_RUN_EVENT_TYPE_COMPLETED = "WORKFLOW_RUN_EVENT_TYPE_COMPLETED"
    WORKFLOW_RUN_EVENT_TYPE_FAILED = "WORKFLOW_RUN_EVENT_TYPE_FAILED"
    WORKFLOW_RUN_EVENT_TYPE_CANCELLED = "WORKFLOW_RUN_EVENT_TYPE_CANCELLED"
    WORKFLOW_RUN_EVENT_TYPE_TIMED_OUT = "WORKFLOW_RUN_EVENT_TYPE_TIMED_OUT"


step_run_event_type_mapping = {
    ResourceEventType.RESOURCE_EVENT_TYPE_STARTED: StepRunEventType.STEP_RUN_EVENT_TYPE_STARTED,
    ResourceEventType.RESOURCE_EVENT_TYPE_COMPLETED: StepRunEventType.STEP_RUN_EVENT_TYPE_COMPLETED,
    ResourceEventType.RESOURCE_EVENT_TYPE_FAILED: StepRunEventType.STEP_RUN_EVENT_TYPE_FAILED,
    ResourceEventType.RESOURCE_EVENT_TYPE_CANCELLED: StepRunEventType.STEP_RUN_EVENT_TYPE_CANCELLED,
    ResourceEventType.RESOURCE_EVENT_TYPE_TIMED_OUT: StepRunEventType.STEP_RUN_EVENT_TYPE_TIMED_OUT,
    ResourceEventType.RESOURCE_EVENT_TYPE_STREAM: StepRunEventType.STEP_RUN_EVENT_TYPE_STREAM,
}

workflow_run_event_type_mapping = {
    ResourceEventType.RESOURCE_EVENT_TYPE_STARTED: WorkflowRunEventType.WORKFLOW_RUN_EVENT_TYPE_STARTED,
    ResourceEventType.RESOURCE_EVENT_TYPE_COMPLETED: WorkflowRunEventType.WORKFLOW_RUN_EVENT_TYPE_COMPLETED,
    ResourceEventType.RESOURCE_EVENT_TYPE_FAILED: WorkflowRunEventType.WORKFLOW_RUN_EVENT_TYPE_FAILED,
    ResourceEventType.RESOURCE_EVENT_TYPE_CANCELLED: WorkflowRunEventType.WORKFLOW_RUN_EVENT_TYPE_CANCELLED,
    ResourceEventType.RESOURCE_EVENT_TYPE_TIMED_OUT: WorkflowRunEventType.WORKFLOW_RUN_EVENT_TYPE_TIMED_OUT,
}

T = TypeVar("T")


class StepRunEvent(BaseModel):
    type: StepRunEventType
    payload: str


class RunEventListener:
    def __init__(
        self,
        config: ClientConfig,
        workflow_run_id: str | None = None,
        additional_meta_kv: tuple[str, str] | None = None,
    ):
        self.config = config
        self.stop_signal = False

        self.workflow_run_id = workflow_run_id
        self.additional_meta_kv = additional_meta_kv

        ## IMPORTANT: This needs to be created lazily so we don't require
        ## an event loop to instantiate the client.
        self.client: DispatcherStub | None = None

    def abort(self) -> None:
        self.stop_signal = True

    def __aiter__(self) -> AsyncGenerator[StepRunEvent, None]:
        return self._generator()

    async def __anext__(self) -> StepRunEvent:
        return await self._generator().__anext__()

    def async_to_sync_thread(
        self, async_iter: AsyncGenerator[T, None]
    ) -> Generator[T, None, None]:
        q = Queue[T | Literal["DONE"]]()
        done_sentinel: Literal["DONE"] = "DONE"

        def runner() -> None:
            loop = asyncio.new_event_loop()
            asyncio.set_event_loop(loop)

            async def consume() -> None:
                try:
                    async for item in async_iter:
                        q.put(item)
                finally:
                    q.put(done_sentinel)

            try:
                loop.run_until_complete(consume())
            finally:
                loop.stop()
                loop.close()

        thread = Thread(target=runner)
        thread.start()

        while True:
            try:
                item = q.get(timeout=1)
                if item == "DONE":
                    break
                yield item
            except Empty:
                continue

        thread.join()

    def __iter__(self) -> Generator[StepRunEvent, None, None]:
        for item in self.async_to_sync_thread(self.__aiter__()):
            yield item

    async def _generator(self) -> AsyncGenerator[StepRunEvent, None]:
        while True:
            if self.stop_signal:
                listener = None
                break

            listener = await self.retry_subscribe()

            try:
                async for workflow_event in listener:
                    eventType = None
                    if workflow_event.resourceType == RESOURCE_TYPE_STEP_RUN:
                        if workflow_event.eventType in step_run_event_type_mapping:
                            eventType = step_run_event_type_mapping[
                                workflow_event.eventType
                            ]
                        else:
                            raise Exception(
                                f"Unknown event type: {workflow_event.eventType}"
                            )

                        yield StepRunEvent(
                            type=eventType, payload=workflow_event.eventPayload
                        )
                    elif workflow_event.resourceType == RESOURCE_TYPE_WORKFLOW_RUN:
                        if workflow_event.eventType in step_run_event_type_mapping:
                            workflowRunEventType = step_run_event_type_mapping[
                                workflow_event.eventType
                            ]
                        else:
                            raise Exception(
                                f"Unknown event type: {workflow_event.eventType}"
                            )

                        yield StepRunEvent(
                            type=workflowRunEventType,
                            payload=workflow_event.eventPayload,
                        )

                    if workflow_event.hangup:
                        listener = None
                        break

                break
            except grpc.RpcError as e:
                # Handle different types of errors
                if e.code() == grpc.StatusCode.CANCELLED:
                    # Context cancelled, unsubscribe and close
                    break
                elif e.code() == grpc.StatusCode.UNAVAILABLE:
                    # Retry logic
                    # logger.info("Could not connect to Hatchet, retrying...")
                    listener = await self.retry_subscribe()
                elif e.code() == grpc.StatusCode.DEADLINE_EXCEEDED:
                    # logger.info("Deadline exceeded, retrying subscription")
                    continue
                else:
                    # Unknown error, report and break
                    # logger.error(f"Failed to receive message: {e}")
                    break
                # Raise StopAsyncIteration to properly end the generator

    async def retry_subscribe(self) -> AsyncGenerator[WorkflowEvent, None]:
        retries = 0

        if self.client is None:
            aio_conn = new_conn(self.config, True)
            self.client = DispatcherStub(aio_conn)

        while retries < DEFAULT_ACTION_LISTENER_RETRY_COUNT:
            try:
                if retries > 0:
                    await asyncio.sleep(DEFAULT_ACTION_LISTENER_RETRY_INTERVAL)

                if self.workflow_run_id is not None:
                    return cast(
                        AsyncGenerator[WorkflowEvent, None],
                        self.client.SubscribeToWorkflowEvents(
                            SubscribeToWorkflowEventsRequest(
                                workflowRunId=self.workflow_run_id,
                            ),
                            metadata=get_metadata(self.config.token),
                        ),
                    )
                elif self.additional_meta_kv is not None:
                    return cast(
                        AsyncGenerator[WorkflowEvent, None],
                        self.client.SubscribeToWorkflowEvents(
                            SubscribeToWorkflowEventsRequest(
                                additionalMetaKey=self.additional_meta_kv[0],
                                additionalMetaValue=self.additional_meta_kv[1],
                            ),
                            metadata=get_metadata(self.config.token),
                        ),
                    )
                else:
                    raise Exception("no listener method provided")

            except grpc.RpcError as e:
                if e.code() == grpc.StatusCode.UNAVAILABLE:
                    retries = retries + 1
                else:
                    raise ValueError(f"gRPC error: {e}")

        raise Exception("Failed to subscribe to workflow events")


class RunEventListenerClient:
    def __init__(self, config: ClientConfig):
        self.config = config

    def stream_by_run_id(self, workflow_run_id: str) -> RunEventListener:
        return self.stream(workflow_run_id)

    def stream(self, workflow_run_id: str) -> RunEventListener:
        return RunEventListener(config=self.config, workflow_run_id=workflow_run_id)

    def stream_by_additional_metadata(self, key: str, value: str) -> RunEventListener:
        return RunEventListener(config=self.config, additional_meta_kv=(key, value))

    async def on(
        self, workflow_run_id: str, handler: Callable[[StepRunEvent], Any] | None = None
    ) -> None:
        async for event in self.stream(workflow_run_id):
            # call the handler if provided
            if handler:
                handler(event)



from enum import Enum
from typing import Type, TypeVar, overload

from google.protobuf.internal.enum_type_wrapper import EnumTypeWrapper

TProtoEnumValue = TypeVar("TProtoEnumValue", bound=int)

TProtoEnum = TypeVar("TProtoEnum", bound=EnumTypeWrapper)
TPythonEnum = TypeVar("TPythonEnum", bound=Enum)


def convert_python_enum_to_proto(
    value: TPythonEnum | None, proto_enum: TProtoEnum
) -> int | None:
    if value is None:
        return None

    names = [item.name for item in proto_enum.DESCRIPTOR.values]

    for name in names:
        if name == value.name:
            return proto_enum.Value(value.name)

    raise ValueError(f"Value must be one of {names}. Got: {value}")


@overload
def convert_proto_enum_to_python(
    value: TProtoEnumValue, python_enum_class: Type[TPythonEnum], proto_enum: TProtoEnum
) -> TPythonEnum: ...


@overload
def convert_proto_enum_to_python(
    value: None, python_enum_class: Type[TPythonEnum], proto_enum: TProtoEnum
) -> None: ...


def convert_proto_enum_to_python(
    value: TProtoEnumValue | None,
    python_enum_class: Type[TPythonEnum],
    proto_enum: TProtoEnum,
) -> TPythonEnum | None:
    if value is None:
        return None

    return python_enum_class[proto_enum.Name(value)]



from importlib.metadata import version
from typing import Any, Callable, Collection, Coroutine

from hatchet_sdk.utils.typing import JSONSerializableMapping

try:
    from opentelemetry.context import Context
    from opentelemetry.instrumentation.instrumentor import (  # type: ignore[attr-defined]
        BaseInstrumentor,
    )
    from opentelemetry.instrumentation.utils import unwrap
    from opentelemetry.metrics import MeterProvider, NoOpMeterProvider, get_meter
    from opentelemetry.trace import (
        NoOpTracerProvider,
        StatusCode,
        TracerProvider,
        get_tracer,
        get_tracer_provider,
    )
    from opentelemetry.trace.propagation.tracecontext import (
        TraceContextTextMapPropagator,
    )
    from wrapt import wrap_function_wrapper  # type: ignore[import-untyped]
except (RuntimeError, ImportError, ModuleNotFoundError):
    raise ModuleNotFoundError(
        "To use the HatchetInstrumentor, you must install Hatchet's `otel` extra using (e.g.) `pip install hatchet-sdk[otel]`"
    )

import hatchet_sdk
from hatchet_sdk.clients.admin import (
    AdminClient,
    TriggerWorkflowOptions,
    WorkflowRunTriggerConfig,
)
from hatchet_sdk.clients.dispatcher.action_listener import Action
from hatchet_sdk.clients.events import (
    BulkPushEventWithMetadata,
    EventClient,
    PushEventOptions,
)
from hatchet_sdk.contracts.events_pb2 import Event
from hatchet_sdk.worker.runner.runner import Runner
from hatchet_sdk.workflow_run import WorkflowRunRef

hatchet_sdk_version = version("hatchet-sdk")

InstrumentKwargs = TracerProvider | MeterProvider | None

OTEL_TRACEPARENT_KEY = "traceparent"


def create_traceparent() -> str | None:
    """
    Creates and returns a W3C traceparent header value using OpenTelemetry's context propagation.

    The traceparent header is used to propagate context information across service boundaries
    in distributed tracing systems. It follows the W3C Trace Context specification.

    :returns: A W3C-formatted traceparent header value if successful, None if the context
                    injection fails or no active span exists.\n
                    Example: `00-4bf92f3577b34da6a3ce929d0e0e4736-00f067aa0ba902b7-01`
    :rtype: str | None:
    """

    carrier: dict[str, str] = {}
    TraceContextTextMapPropagator().inject(carrier)

    return carrier.get("traceparent")


def parse_carrier_from_metadata(
    metadata: JSONSerializableMapping | None,
) -> Context | None:
    """
    Parses OpenTelemetry trace context from a metadata dictionary.

    Extracts the trace context from metadata using the W3C Trace Context format,
    specifically looking for the `traceparent` header.

    :param metadata: A dictionary containing metadata key-value pairs,
                     potentially including the `traceparent` header. Can be None.
    :type metadata: dict[str, str] | None
    :returns: The extracted OpenTelemetry Context object if a valid `traceparent`
              is found in the metadata, otherwise None.
    :rtype: Context | None

    :Example:

    >>> metadata = {"traceparent": "00-0af7651916cd43dd8448eb211c80319c-b7ad6b7169203331-01"}
    >>> context = parse_carrier_from_metadata(metadata)
    """

    if not metadata:
        return None

    traceparent = metadata.get(OTEL_TRACEPARENT_KEY)

    if not traceparent:
        return None

    return TraceContextTextMapPropagator().extract({OTEL_TRACEPARENT_KEY: traceparent})


def inject_traceparent_into_metadata(
    metadata: dict[str, str], traceparent: str | None = None
) -> dict[str, str]:
    """
    Injects OpenTelemetry `traceparent` into a metadata dictionary.

    Takes a metadata dictionary and an optional `traceparent` string,
    returning a new metadata dictionary with the `traceparent` added under the
    `OTEL_TRACEPARENT_KEY`. If no `traceparent` is provided, it attempts to create one.

    :param metadata: The metadata dictionary to inject the `traceparent` into.
    :type metadata: dict[str, str]
    :param traceparent: The `traceparent` string to inject. If None, attempts to use
                        the current span.
    :type traceparent: str | None, optional
    :returns: A new metadata dictionary containing the original metadata plus
              the injected `traceparent`, if one was available or could be created.
    :rtype: dict[str, str]

    :Example:

    >>> metadata = {"key": "value"}
    >>> new_metadata = inject_traceparent(metadata, "00-4bf92f3577b34da6a3ce929d0e0e4736-00f067aa0ba902b7-01")
    >>> print(new_metadata)
    {"key": "value", "traceparent": "00-4bf92f3577b34da6a3ce929d0e0e4736-00f067aa0ba902b7-01"}
    """

    if not traceparent:
        traceparent = create_traceparent()

    if not traceparent:
        return metadata

    return {
        **metadata,
        OTEL_TRACEPARENT_KEY: traceparent,
    }


class HatchetInstrumentor(BaseInstrumentor):  # type: ignore[misc]
    def __init__(
        self,
        tracer_provider: TracerProvider | None = None,
        meter_provider: MeterProvider | None = None,
    ):
        """
        Hatchet OpenTelemetry instrumentor.

        The instrumentor provides an OpenTelemetry integration for Hatchet by setting up
        tracing and metrics collection.

        :param tracer_provider: TracerProvider | None: The OpenTelemetry TracerProvider to use.
                If not provided, the global tracer provider will be used.
        :param meter_provider: MeterProvider | None: The OpenTelemetry MeterProvider to use.
                If not provided, a no-op meter provider will be used.
        """

        self.tracer_provider = tracer_provider or get_tracer_provider()
        self.meter_provider = meter_provider or NoOpMeterProvider()

        super().__init__()

    def instrumentation_dependencies(self) -> Collection[str]:
        return tuple()

    def _instrument(self, **kwargs: InstrumentKwargs) -> None:
        self._tracer = get_tracer(__name__, hatchet_sdk_version, self.tracer_provider)
        self._meter = get_meter(__name__, hatchet_sdk_version, self.meter_provider)

        wrap_function_wrapper(
            hatchet_sdk,
            "worker.runner.runner.Runner.handle_start_step_run",
            self._wrap_handle_start_step_run,
        )
        wrap_function_wrapper(
            hatchet_sdk,
            "worker.runner.runner.Runner.handle_start_group_key_run",
            self._wrap_handle_get_group_key_run,
        )
        wrap_function_wrapper(
            hatchet_sdk,
            "worker.runner.runner.Runner.handle_cancel_action",
            self._wrap_handle_cancel_action,
        )

        wrap_function_wrapper(
            hatchet_sdk,
            "clients.events.EventClient.push",
            self._wrap_push_event,
        )

        wrap_function_wrapper(
            hatchet_sdk,
            "clients.events.EventClient.bulk_push",
            self._wrap_bulk_push_event,
        )

        wrap_function_wrapper(
            hatchet_sdk,
            "clients.admin.AdminClient.run_workflow",
            self._wrap_run_workflow,
        )

        wrap_function_wrapper(
            hatchet_sdk,
            "clients.admin.AdminClient.aio_run_workflow",
            self._wrap_async_run_workflow,
        )

        wrap_function_wrapper(
            hatchet_sdk,
            "clients.admin.AdminClient.run_workflows",
            self._wrap_run_workflows,
        )

        wrap_function_wrapper(
            hatchet_sdk,
            "clients.admin.AdminClient.aio_run_workflows",
            self._wrap_async_run_workflows,
        )

    ## IMPORTANT: Keep these types in sync with the wrapped method's signature
    async def _wrap_handle_start_step_run(
        self,
        wrapped: Callable[[Action], Coroutine[None, None, Exception | None]],
        instance: Runner,
        args: tuple[Action],
        kwargs: Any,
    ) -> Exception | None:
        action = args[0]
        traceparent = parse_carrier_from_metadata(action.additional_metadata)

        with self._tracer.start_as_current_span(
            "hatchet.start_step_run",
            attributes=action.otel_attributes,
            context=traceparent,
        ) as span:
            result = await wrapped(*args, **kwargs)

            if isinstance(result, Exception):
                span.set_status(StatusCode.ERROR, str(result))

            return result

    ## IMPORTANT: Keep these types in sync with the wrapped method's signature
    async def _wrap_handle_get_group_key_run(
        self,
        wrapped: Callable[[Action], Coroutine[None, None, Exception | None]],
        instance: Runner,
        args: tuple[Action],
        kwargs: Any,
    ) -> Exception | None:
        action = args[0]

        with self._tracer.start_as_current_span(
            "hatchet.get_group_key_run",
            attributes=action.otel_attributes,
        ) as span:
            result = await wrapped(*args, **kwargs)

            if isinstance(result, Exception):
                span.set_status(StatusCode.ERROR, str(result))

            return result

    ## IMPORTANT: Keep these types in sync with the wrapped method's signature
    async def _wrap_handle_cancel_action(
        self,
        wrapped: Callable[[str], Coroutine[None, None, Exception | None]],
        instance: Runner,
        args: tuple[str],
        kwargs: Any,
    ) -> Exception | None:
        step_run_id = args[0]

        with self._tracer.start_as_current_span(
            "hatchet.cancel_step_run",
            attributes={
                "hatchet.step_run_id": step_run_id,
            },
        ):
            return await wrapped(*args, **kwargs)

    ## IMPORTANT: Keep these types in sync with the wrapped method's signature
    def _wrap_push_event(
        self,
        wrapped: Callable[[str, dict[str, Any], PushEventOptions | None], Event],
        instance: EventClient,
        args: tuple[
            str,
            dict[str, Any],
            PushEventOptions | None,
        ],
        kwargs: dict[str, str | dict[str, Any] | PushEventOptions | None],
    ) -> Event:
        with self._tracer.start_as_current_span(
            "hatchet.push_event",
        ):
            return wrapped(*args, **kwargs)

    ## IMPORTANT: Keep these types in sync with the wrapped method's signature
    def _wrap_bulk_push_event(
        self,
        wrapped: Callable[
            [list[BulkPushEventWithMetadata], PushEventOptions | None], list[Event]
        ],
        instance: EventClient,
        args: tuple[
            list[BulkPushEventWithMetadata],
            PushEventOptions | None,
        ],
        kwargs: dict[str, list[BulkPushEventWithMetadata] | PushEventOptions | None],
    ) -> list[Event]:
        with self._tracer.start_as_current_span(
            "hatchet.bulk_push_event",
        ):
            return wrapped(*args, **kwargs)

    ## IMPORTANT: Keep these types in sync with the wrapped method's signature
    def _wrap_run_workflow(
        self,
        wrapped: Callable[[str, Any, TriggerWorkflowOptions | None], WorkflowRunRef],
        instance: AdminClient,
        args: tuple[str, Any, TriggerWorkflowOptions | None],
        kwargs: dict[str, str | Any | TriggerWorkflowOptions | None],
    ) -> WorkflowRunRef:
        with self._tracer.start_as_current_span(
            "hatchet.run_workflow",
        ):
            return wrapped(*args, **kwargs)

    ## IMPORTANT: Keep these types in sync with the wrapped method's signature
    async def _wrap_async_run_workflow(
        self,
        wrapped: Callable[
            [str, Any, TriggerWorkflowOptions | None],
            Coroutine[None, None, WorkflowRunRef],
        ],
        instance: AdminClient,
        args: tuple[str, Any, TriggerWorkflowOptions | None],
        kwargs: dict[str, str | Any | TriggerWorkflowOptions | None],
    ) -> WorkflowRunRef:
        with self._tracer.start_as_current_span(
            "hatchet.run_workflow",
        ):
            return await wrapped(*args, **kwargs)

    ## IMPORTANT: Keep these types in sync with the wrapped method's signature
    def _wrap_run_workflows(
        self,
        wrapped: Callable[
            [list[WorkflowRunTriggerConfig]],
            list[WorkflowRunRef],
        ],
        instance: AdminClient,
        args: tuple[list[WorkflowRunTriggerConfig],],
        kwargs: dict[str, list[WorkflowRunTriggerConfig]],
    ) -> list[WorkflowRunRef]:
        with self._tracer.start_as_current_span(
            "hatchet.run_workflows",
        ):
            return wrapped(*args, **kwargs)

    ## IMPORTANT: Keep these types in sync with the wrapped method's signature
    async def _wrap_async_run_workflows(
        self,
        wrapped: Callable[
            [list[WorkflowRunTriggerConfig]],
            Coroutine[None, None, list[WorkflowRunRef]],
        ],
        instance: AdminClient,
        args: tuple[list[WorkflowRunTriggerConfig],],
        kwargs: dict[str, list[WorkflowRunTriggerConfig]],
    ) -> list[WorkflowRunRef]:
        with self._tracer.start_as_current_span(
            "hatchet.run_workflows",
        ):
            return await wrapped(*args, **kwargs)

    def _uninstrument(self, **kwargs: InstrumentKwargs) -> None:
        self.tracer_provider = NoOpTracerProvider()
        self.meter_provider = NoOpMeterProvider()

        unwrap(hatchet_sdk, "worker.runner.runner.Runner.handle_start_step_run")
        unwrap(hatchet_sdk, "worker.runner.runner.Runner.handle_start_group_key_run")
        unwrap(hatchet_sdk, "worker.runner.runner.Runner.handle_cancel_action")
        unwrap(hatchet_sdk, "clients.events.EventClient.push")
        unwrap(hatchet_sdk, "clients.events.EventClient.bulk_push")
        unwrap(hatchet_sdk, "clients.admin.AdminClient.run_workflow")
        unwrap(hatchet_sdk, "clients.admin.AdminClient.aio_run_workflow")
        unwrap(hatchet_sdk, "clients.admin.AdminClient.run_workflows")
        unwrap(hatchet_sdk, "clients.admin.AdminClient.aio_run_workflows")



import asyncio
import datetime
from typing import Optional

from hatchet_sdk.clients.rest.api.workflow_api import WorkflowApi
from hatchet_sdk.clients.rest.api.workflow_run_api import WorkflowRunApi
from hatchet_sdk.clients.rest.api_client import ApiClient
from hatchet_sdk.clients.rest.models.schedule_workflow_run_request import (
    ScheduleWorkflowRunRequest,
)
from hatchet_sdk.clients.rest.models.scheduled_run_status import ScheduledRunStatus
from hatchet_sdk.clients.rest.models.scheduled_workflows import ScheduledWorkflows
from hatchet_sdk.clients.rest.models.scheduled_workflows_list import (
    ScheduledWorkflowsList,
)
from hatchet_sdk.clients.rest.models.scheduled_workflows_order_by_field import (
    ScheduledWorkflowsOrderByField,
)
from hatchet_sdk.clients.rest.models.workflow_run_order_by_direction import (
    WorkflowRunOrderByDirection,
)
from hatchet_sdk.clients.v1.api_client import (
    BaseRestClient,
    maybe_additional_metadata_to_kv,
)
from hatchet_sdk.utils.typing import JSONSerializableMapping


class ScheduledClient(BaseRestClient):
    def _wra(self, client: ApiClient) -> WorkflowRunApi:
        return WorkflowRunApi(client)

    def _wa(self, client: ApiClient) -> WorkflowApi:
        return WorkflowApi(client)

    def create(
        self,
        workflow_name: str,
        trigger_at: datetime.datetime,
        input: JSONSerializableMapping,
        additional_metadata: JSONSerializableMapping,
    ) -> ScheduledWorkflows:
        """
        Creates a new scheduled workflow run asynchronously.

        Args:
            workflow_name (str): The name of the scheduled workflow.
            trigger_at (datetime.datetime): The datetime when the run should be triggered.
            input (JSONSerializableMapping): The input data for the scheduled workflow.
            additional_metadata (JSONSerializableMapping): Additional metadata associated with the future run.

        Returns:
            ScheduledWorkflows: The created scheduled workflow instance.
        """
        with self.client() as client:
            return self._wra(client).scheduled_workflow_run_create(
                tenant=self.client_config.tenant_id,
                workflow=workflow_name,
                schedule_workflow_run_request=ScheduleWorkflowRunRequest(
                    triggerAt=trigger_at,
                    input=dict(input),
                    additionalMetadata=dict(additional_metadata),
                ),
            )

    async def aio_create(
        self,
        workflow_name: str,
        trigger_at: datetime.datetime,
        input: JSONSerializableMapping,
        additional_metadata: JSONSerializableMapping,
    ) -> ScheduledWorkflows:
        """
        Creates a new scheduled workflow run asynchronously.

        Args:
            workflow_name (str): The name of the scheduled workflow.
            trigger_at (datetime.datetime): The datetime when the run should be triggered.
            input (JSONSerializableMapping): The input data for the scheduled workflow.
            additional_metadata (JSONSerializableMapping): Additional metadata associated with the future run as a key-value pair (e.g. {"key1": "value1", "key2": "value2"}).

        Returns:
            ScheduledWorkflows: The created scheduled workflow instance.
        """

        return await asyncio.to_thread(
            self.create,
            workflow_name,
            trigger_at,
            input,
            additional_metadata,
        )

    def delete(self, scheduled_id: str) -> None:
        """
        Deletes a scheduled workflow run.

        Args:
            scheduled_id (str): The scheduled workflow trigger ID to delete.
        """
        with self.client() as client:
            self._wa(client).workflow_scheduled_delete(
                tenant=self.client_config.tenant_id,
                scheduled_workflow_run=scheduled_id,
            )

    async def aio_delete(self, scheduled_id: str) -> None:
        await asyncio.to_thread(self.delete, scheduled_id)

    async def aio_list(
        self,
        offset: int | None = None,
        limit: int | None = None,
        workflow_id: str | None = None,
        parent_workflow_run_id: str | None = None,
        statuses: list[ScheduledRunStatus] | None = None,
        additional_metadata: Optional[JSONSerializableMapping] = None,
        order_by_field: Optional[ScheduledWorkflowsOrderByField] = None,
        order_by_direction: Optional[WorkflowRunOrderByDirection] = None,
    ) -> ScheduledWorkflowsList:
        """
        Retrieves a list of scheduled workflows based on provided filters.

        Args:
            offset (int | None): The starting point for the list.
            limit (int | None): The maximum number of items to return.
            workflow_id (str | None): Filter by specific workflow ID.
            parent_workflow_run_id (str | None): Filter by parent workflow run ID.
            statuses (list[ScheduledRunStatus] | None): Filter by status.
            additional_metadata (Optional[List[dict[str, str]]]): Filter by additional metadata.
            order_by_field (Optional[ScheduledWorkflowsOrderByField]): Field to order the results by.
            order_by_direction (Optional[WorkflowRunOrderByDirection]): Direction to order the results.

        Returns:
            List[ScheduledWorkflows]: A list of scheduled workflows matching the criteria.
        """
        return await asyncio.to_thread(
            self.list,
            offset=offset,
            limit=limit,
            workflow_id=workflow_id,
            additional_metadata=additional_metadata,
            order_by_field=order_by_field,
            order_by_direction=order_by_direction,
            parent_workflow_run_id=parent_workflow_run_id,
            statuses=statuses,
        )

    def list(
        self,
        offset: int | None = None,
        limit: int | None = None,
        workflow_id: str | None = None,
        parent_workflow_run_id: str | None = None,
        statuses: list[ScheduledRunStatus] | None = None,
        additional_metadata: Optional[JSONSerializableMapping] = None,
        order_by_field: Optional[ScheduledWorkflowsOrderByField] = None,
        order_by_direction: Optional[WorkflowRunOrderByDirection] = None,
    ) -> ScheduledWorkflowsList:
        """
        Retrieves a list of scheduled workflows based on provided filters.

        Args:
            offset (int | None): The starting point for the list.
            limit (int | None): The maximum number of items to return.
            workflow_id (str | None): Filter by specific workflow ID.
            parent_workflow_run_id (str | None): Filter by parent workflow run ID.
            statuses (list[ScheduledRunStatus] | None): Filter by status.
            additional_metadata (Optional[List[dict[str, str]]]): Filter by additional metadata.
            order_by_field (Optional[ScheduledWorkflowsOrderByField]): Field to order the results by.
            order_by_direction (Optional[WorkflowRunOrderByDirection]): Direction to order the results.

        Returns:
            List[ScheduledWorkflows]: A list of scheduled workflows matching the criteria.
        """
        with self.client() as client:
            return self._wa(client).workflow_scheduled_list(
                tenant=self.client_config.tenant_id,
                offset=offset,
                limit=limit,
                order_by_field=order_by_field,
                order_by_direction=order_by_direction,
                workflow_id=workflow_id,
                additional_metadata=maybe_additional_metadata_to_kv(
                    additional_metadata
                ),
                parent_workflow_run_id=parent_workflow_run_id,
                statuses=statuses,
            )

    def get(self, scheduled_id: str) -> ScheduledWorkflows:
        """
        Retrieves a specific scheduled workflow by scheduled run trigger ID.

        Args:
            scheduled (str): The scheduled workflow trigger ID to retrieve.

        Returns:
            ScheduledWorkflows: The requested scheduled workflow instance.
        """

        with self.client() as client:
            return self._wa(client).workflow_scheduled_get(
                tenant=self.client_config.tenant_id,
                scheduled_workflow_run=scheduled_id,
            )

    async def aio_get(self, scheduled_id: str) -> ScheduledWorkflows:
        """
        Retrieves a specific scheduled workflow by scheduled run trigger ID.

        Args:
            scheduled (str): The scheduled workflow trigger ID to retrieve.

        Returns:
            ScheduledWorkflows: The requested scheduled workflow instance.
        """
        return await asyncio.to_thread(self.get, scheduled_id)



import asyncio
import json
import time
from dataclasses import field
from enum import Enum
from typing import Any, AsyncGenerator, cast

import grpc
import grpc.aio
from grpc._cython import cygrpc  # type: ignore[attr-defined]
from pydantic import BaseModel, ConfigDict, Field, field_validator, model_validator

from hatchet_sdk.clients.event_ts import ThreadSafeEvent, read_with_interrupt
from hatchet_sdk.clients.events import proto_timestamp_now
from hatchet_sdk.clients.listeners.run_event_listener import (
    DEFAULT_ACTION_LISTENER_RETRY_INTERVAL,
)
from hatchet_sdk.config import ClientConfig
from hatchet_sdk.connection import new_conn
from hatchet_sdk.contracts.dispatcher_pb2 import ActionType as ActionTypeProto
from hatchet_sdk.contracts.dispatcher_pb2 import (
    AssignedAction,
    HeartbeatRequest,
    WorkerLabels,
    WorkerListenRequest,
    WorkerUnsubscribeRequest,
)
from hatchet_sdk.contracts.dispatcher_pb2_grpc import DispatcherStub
from hatchet_sdk.logger import logger
from hatchet_sdk.metadata import get_metadata
from hatchet_sdk.utils.backoff import exp_backoff_sleep
from hatchet_sdk.utils.proto_enums import convert_proto_enum_to_python
from hatchet_sdk.utils.typing import JSONSerializableMapping

DEFAULT_ACTION_TIMEOUT = 600  # seconds
DEFAULT_ACTION_LISTENER_RETRY_COUNT = 15


class GetActionListenerRequest(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True)

    worker_name: str
    services: list[str]
    actions: list[str]
    slots: int
    raw_labels: dict[str, str | int] = Field(default_factory=dict)

    labels: dict[str, WorkerLabels] = Field(default_factory=dict)

    @model_validator(mode="after")
    def validate_labels(self) -> "GetActionListenerRequest":
        self.labels = {}

        for key, value in self.raw_labels.items():
            if isinstance(value, int):
                self.labels[key] = WorkerLabels(intValue=value)
            else:
                self.labels[key] = WorkerLabels(strValue=str(value))

        return self


class ActionPayload(BaseModel):
    model_config = ConfigDict(extra="allow")

    input: JSONSerializableMapping = Field(default_factory=dict)
    parents: dict[str, JSONSerializableMapping] = Field(default_factory=dict)
    overrides: JSONSerializableMapping = Field(default_factory=dict)
    user_data: JSONSerializableMapping = Field(default_factory=dict)
    step_run_errors: dict[str, str] = Field(default_factory=dict)
    triggered_by: str | None = None
    triggers: JSONSerializableMapping = Field(default_factory=dict)

    @field_validator(
        "input", "parents", "overrides", "user_data", "step_run_errors", mode="before"
    )
    @classmethod
    def validate_fields(cls, v: Any) -> Any:
        return v or {}


class ActionType(str, Enum):
    START_STEP_RUN = "START_STEP_RUN"
    CANCEL_STEP_RUN = "CANCEL_STEP_RUN"
    START_GET_GROUP_KEY = "START_GET_GROUP_KEY"


class Action(BaseModel):
    worker_id: str
    tenant_id: str
    workflow_run_id: str
    get_group_key_run_id: str
    job_id: str
    job_name: str
    job_run_id: str
    step_id: str
    step_run_id: str
    action_id: str
    action_type: ActionType
    retry_count: int
    action_payload: ActionPayload
    additional_metadata: JSONSerializableMapping = field(default_factory=dict)

    child_workflow_index: int | None = None
    child_workflow_key: str | None = None
    parent_workflow_run_id: str | None = None

    def _dump_payload_to_str(self) -> str:
        try:
            return json.dumps(self.action_payload.model_dump(), default=str)
        except Exception:
            return str(self.action_payload)

    @property
    def otel_attributes(self) -> dict[str, str | int]:
        try:
            payload_str = json.dumps(self.action_payload.model_dump(), default=str)
        except Exception:
            payload_str = str(self.action_payload)

        attrs: dict[str, str | int | None] = {
            "hatchet.tenant_id": self.tenant_id,
            "hatchet.worker_id": self.worker_id,
            "hatchet.workflow_run_id": self.workflow_run_id,
            "hatchet.step_id": self.step_id,
            "hatchet.step_run_id": self.step_run_id,
            "hatchet.retry_count": self.retry_count,
            "hatchet.parent_workflow_run_id": self.parent_workflow_run_id,
            "hatchet.child_workflow_index": self.child_workflow_index,
            "hatchet.child_workflow_key": self.child_workflow_key,
            "hatchet.action_payload": payload_str,
            "hatchet.workflow_name": self.job_name,
            "hatchet.action_name": self.action_id,
            "hatchet.get_group_key_run_id": self.get_group_key_run_id,
        }

        return {k: v for k, v in attrs.items() if v}


def parse_additional_metadata(additional_metadata: str) -> JSONSerializableMapping:
    try:
        return cast(
            JSONSerializableMapping,
            json.loads(additional_metadata),
        )
    except json.JSONDecodeError:
        return {}


class ActionListener:
    def __init__(self, config: ClientConfig, worker_id: str) -> None:
        self.config = config
        self.worker_id = worker_id

        self.aio_client = DispatcherStub(new_conn(self.config, True))
        self.token = self.config.token

        self.retries = 0
        self.last_heartbeat_succeeded = True
        self.time_last_hb_succeeded = 9999999999999.0
        self.last_connection_attempt = 0.0
        self.heartbeat_task: asyncio.Task[None] | None = None
        self.run_heartbeat = True
        self.listen_strategy = "v2"
        self.stop_signal = False
        self.missed_heartbeats = 0

    def is_healthy(self) -> bool:
        return self.last_heartbeat_succeeded

    async def heartbeat(self) -> None:
        # send a heartbeat every 4 seconds
        heartbeat_delay = 4

        while True:
            if not self.run_heartbeat:
                break

            try:
                logger.debug("sending heartbeat")
                await self.aio_client.Heartbeat(
                    HeartbeatRequest(
                        workerId=self.worker_id,
                        heartbeatAt=proto_timestamp_now(),
                    ),
                    timeout=5,
                    metadata=get_metadata(self.token),
                )

                if self.last_heartbeat_succeeded is False:
                    logger.info("listener established")

                now = time.time()
                diff = now - self.time_last_hb_succeeded
                if diff > heartbeat_delay + 1:
                    logger.warn(
                        f"time since last successful heartbeat: {diff:.2f}s, expects {heartbeat_delay}s"
                    )

                self.last_heartbeat_succeeded = True
                self.time_last_hb_succeeded = now
                self.missed_heartbeats = 0
            except grpc.RpcError as e:
                self.missed_heartbeats = self.missed_heartbeats + 1
                self.last_heartbeat_succeeded = False

                if (
                    e.code() == grpc.StatusCode.UNAVAILABLE
                    or e.code() == grpc.StatusCode.FAILED_PRECONDITION
                ):
                    # todo case on "recvmsg:Connection reset by peer" for updates?
                    if self.missed_heartbeats >= 3:
                        # we don't reraise the error here, as we don't want to stop the heartbeat thread
                        logger.error(
                            f"⛔️ failed heartbeat ({self.missed_heartbeats}): {e.details()}"
                        )
                    elif self.missed_heartbeats > 1:
                        logger.warning(
                            f"failed to send heartbeat ({self.missed_heartbeats}): {e.details()}"
                        )
                else:
                    logger.error(f"failed to send heartbeat: {e}")

                if self.interrupt is not None:
                    self.interrupt.set()

                if e.code() == grpc.StatusCode.UNIMPLEMENTED:
                    break
            await asyncio.sleep(heartbeat_delay)

    async def start_heartbeater(self) -> None:
        if self.heartbeat_task is not None:
            return

        loop = asyncio.get_event_loop()

        self.heartbeat_task = loop.create_task(self.heartbeat())

    def __aiter__(self) -> AsyncGenerator[Action | None, None]:
        return self._generator()

    async def _generator(self) -> AsyncGenerator[Action | None, None]:
        listener = None

        while not self.stop_signal:
            if listener is not None:
                listener.cancel()

            try:
                listener = await self.get_listen_client()
            except Exception:
                logger.info("closing action listener loop")
                yield None

            try:
                while not self.stop_signal:
                    self.interrupt = ThreadSafeEvent()

                    if listener is None:
                        continue

                    t = asyncio.create_task(
                        read_with_interrupt(listener, self.interrupt)
                    )
                    await self.interrupt.wait()

                    if not t.done():
                        # print a warning
                        logger.warning(
                            "Interrupted read_with_interrupt task of action listener"
                        )

                        t.cancel()
                        listener.cancel()

                        break

                    assigned_action, _ = t.result()

                    if assigned_action is cygrpc.EOF:
                        self.retries = self.retries + 1
                        break

                    self.retries = 0

                    try:
                        action_payload = (
                            ActionPayload()
                            if not assigned_action.actionPayload
                            else ActionPayload.model_validate_json(
                                assigned_action.actionPayload
                            )
                        )
                    except (ValueError, json.JSONDecodeError) as e:
                        logger.error(f"Error decoding payload: {e}")

                        action_payload = ActionPayload()

                    action = Action(
                        tenant_id=assigned_action.tenantId,
                        worker_id=self.worker_id,
                        workflow_run_id=assigned_action.workflowRunId,
                        get_group_key_run_id=assigned_action.getGroupKeyRunId,
                        job_id=assigned_action.jobId,
                        job_name=assigned_action.jobName,
                        job_run_id=assigned_action.jobRunId,
                        step_id=assigned_action.stepId,
                        step_run_id=assigned_action.stepRunId,
                        action_id=assigned_action.actionId,
                        action_payload=action_payload,
                        action_type=convert_proto_enum_to_python(
                            assigned_action.actionType,
                            ActionType,
                            ActionTypeProto,
                        ),
                        retry_count=assigned_action.retryCount,
                        additional_metadata=parse_additional_metadata(
                            assigned_action.additional_metadata
                        ),
                        child_workflow_index=assigned_action.child_workflow_index,
                        child_workflow_key=assigned_action.child_workflow_key,
                        parent_workflow_run_id=assigned_action.parent_workflow_run_id,
                    )

                    yield action
            except grpc.RpcError as e:
                self.last_heartbeat_succeeded = False

                # Handle different types of errors
                if e.code() == grpc.StatusCode.CANCELLED:
                    # Context cancelled, unsubscribe and close
                    logger.debug("Context cancelled, closing listener")
                elif e.code() == grpc.StatusCode.DEADLINE_EXCEEDED:
                    logger.info("Deadline exceeded, retrying subscription")
                elif (
                    self.listen_strategy == "v2"
                    and e.code() == grpc.StatusCode.UNIMPLEMENTED
                ):
                    # ListenV2 is not available, fallback to Listen
                    self.listen_strategy = "v1"
                    self.run_heartbeat = False
                    logger.info("ListenV2 not available, falling back to Listen")
                else:
                    # TODO retry
                    if e.code() == grpc.StatusCode.UNAVAILABLE:
                        logger.error(f"action listener error: {e.details()}")
                    else:
                        # Unknown error, report and break
                        logger.error(f"action listener error: {e}")

                    self.retries = self.retries + 1

    async def get_listen_client(
        self,
    ) -> grpc.aio.UnaryStreamCall[WorkerListenRequest, AssignedAction]:
        current_time = time.time()

        if (
            current_time - self.last_connection_attempt
            > DEFAULT_ACTION_LISTENER_RETRY_INTERVAL
        ):
            # reset retries if last connection was long lived
            self.retries = 0

        if self.retries > DEFAULT_ACTION_LISTENER_RETRY_COUNT:
            # TODO this is the problem case...
            logger.error(
                f"could not establish action listener connection after {DEFAULT_ACTION_LISTENER_RETRY_COUNT} retries"
            )
            self.run_heartbeat = False
            raise Exception("retry_exhausted")
        elif self.retries >= 1:
            # logger.info
            # if we are retrying, we wait for a bit. this should eventually be replaced with exp backoff + jitter
            await exp_backoff_sleep(
                self.retries, DEFAULT_ACTION_LISTENER_RETRY_INTERVAL
            )

            logger.info(
                f"action listener connection interrupted, retrying... ({self.retries}/{DEFAULT_ACTION_LISTENER_RETRY_COUNT})"
            )

        self.aio_client = DispatcherStub(new_conn(self.config, True))

        if self.listen_strategy == "v2":
            # we should await for the listener to be established before
            # starting the heartbeater
            listener = self.aio_client.ListenV2(
                WorkerListenRequest(workerId=self.worker_id),
                timeout=self.config.listener_v2_timeout,
                metadata=get_metadata(self.token),
            )
            await self.start_heartbeater()
        else:
            # if ListenV2 is not available, fallback to Listen
            listener = self.aio_client.Listen(
                WorkerListenRequest(workerId=self.worker_id),
                timeout=DEFAULT_ACTION_TIMEOUT,
                metadata=get_metadata(self.token),
            )

        self.last_connection_attempt = current_time

        return cast(
            grpc.aio.UnaryStreamCall[WorkerListenRequest, AssignedAction], listener
        )

    def cleanup(self) -> None:
        self.run_heartbeat = False
        if self.heartbeat_task is not None:
            self.heartbeat_task.cancel()

        try:
            self.unregister()
        except Exception as e:
            logger.error(f"failed to unregister: {e}")

        if self.interrupt:  # type: ignore[truthy-bool]
            self.interrupt.set()

    def unregister(self) -> WorkerUnsubscribeRequest:
        self.run_heartbeat = False

        if self.heartbeat_task is not None:
            self.heartbeat_task.cancel()

        try:
            req = self.aio_client.Unsubscribe(
                WorkerUnsubscribeRequest(workerId=self.worker_id),
                timeout=5,
                metadata=get_metadata(self.token),
            )

            if self.interrupt is not None:
                self.interrupt.set()

            return cast(WorkerUnsubscribeRequest, req)
        except grpc.RpcError as e:
            raise Exception(f"Failed to unsubscribe: {e}")



# coding: utf-8

"""
    Hatchet API

    The Hatchet API

    The version of the OpenAPI document: 1.0.0
    Generated by OpenAPI Generator (https://openapi-generator.tech)

    Do not edit the class manually.
"""  # noqa: E501


from __future__ import annotations

import json
import pprint
import re  # noqa: F401
from typing import Any, ClassVar, Dict, List, Optional, Set

from pydantic import BaseModel, ConfigDict, Field, StrictStr
from typing_extensions import Self

from hatchet_sdk.v0.clients.rest.models.api_resource_meta import APIResourceMeta
from hatchet_sdk.v0.clients.rest.models.event_workflow_run_summary import (
    EventWorkflowRunSummary,
)
from hatchet_sdk.v0.clients.rest.models.tenant import Tenant


class Event(BaseModel):
    """
    Event
    """  # noqa: E501

    metadata: APIResourceMeta
    key: StrictStr = Field(description="The key for the event.")
    tenant: Optional[Tenant] = Field(
        default=None, description="The tenant associated with this event."
    )
    tenant_id: StrictStr = Field(
        description="The ID of the tenant associated with this event.", alias="tenantId"
    )
    workflow_run_summary: Optional[EventWorkflowRunSummary] = Field(
        default=None,
        description="The workflow run summary for this event.",
        alias="workflowRunSummary",
    )
    additional_metadata: Optional[Dict[str, Any]] = Field(
        default=None,
        description="Additional metadata for the event.",
        alias="additionalMetadata",
    )
    __properties: ClassVar[List[str]] = [
        "metadata",
        "key",
        "tenant",
        "tenantId",
        "workflowRunSummary",
        "additionalMetadata",
    ]

    model_config = ConfigDict(
        populate_by_name=True,
        validate_assignment=True,
        protected_namespaces=(),
    )

    def to_str(self) -> str:
        """Returns the string representation of the model using alias"""
        return pprint.pformat(self.model_dump(by_alias=True))

    def to_json(self) -> str:
        """Returns the JSON representation of the model using alias"""
        # TODO: pydantic v2: use .model_dump_json(by_alias=True, exclude_unset=True) instead
        return json.dumps(self.to_dict())

    @classmethod
    def from_json(cls, json_str: str) -> Optional[Self]:
        """Create an instance of Event from a JSON string"""
        return cls.from_dict(json.loads(json_str))

    def to_dict(self) -> Dict[str, Any]:
        """Return the dictionary representation of the model using alias.

        This has the following differences from calling pydantic's
        `self.model_dump(by_alias=True)`:

        * `None` is only added to the output dict for nullable fields that
          were set at model initialization. Other fields with value `None`
          are ignored.
        """
        excluded_fields: Set[str] = set([])

        _dict = self.model_dump(
            by_alias=True,
            exclude=excluded_fields,
            exclude_none=True,
        )
        # override the default output from pydantic by calling `to_dict()` of metadata
        if self.metadata:
            _dict["metadata"] = self.metadata.to_dict()
        # override the default output from pydantic by calling `to_dict()` of tenant
        if self.tenant:
            _dict["tenant"] = self.tenant.to_dict()
        # override the default output from pydantic by calling `to_dict()` of workflow_run_summary
        if self.workflow_run_summary:
            _dict["workflowRunSummary"] = self.workflow_run_summary.to_dict()
        return _dict

    @classmethod
    def from_dict(cls, obj: Optional[Dict[str, Any]]) -> Optional[Self]:
        """Create an instance of Event from a dict"""
        if obj is None:
            return None

        if not isinstance(obj, dict):
            return cls.model_validate(obj)

        _obj = cls.model_validate(
            {
                "metadata": (
                    APIResourceMeta.from_dict(obj["metadata"])
                    if obj.get("metadata") is not None
                    else None
                ),
                "key": obj.get("key"),
                "tenant": (
                    Tenant.from_dict(obj["tenant"])
                    if obj.get("tenant") is not None
                    else None
                ),
                "tenantId": obj.get("tenantId"),
                "workflowRunSummary": (
                    EventWorkflowRunSummary.from_dict(obj["workflowRunSummary"])
                    if obj.get("workflowRunSummary") is not None
                    else None
                ),
                "additionalMetadata": obj.get("additionalMetadata"),
            }
        )
        return _obj



from typing import ParamSpec, TypeVar

from hatchet_sdk.clients.rest.api_client import ApiClient
from hatchet_sdk.clients.rest.configuration import Configuration
from hatchet_sdk.config import ClientConfig
from hatchet_sdk.utils.typing import JSONSerializableMapping

## Type variables to use with coroutines.
## See https://stackoverflow.com/questions/73240620/the-right-way-to-type-hint-a-coroutine-function
## Return type
R = TypeVar("R")

## Yield type
Y = TypeVar("Y")

## Send type
S = TypeVar("S")

P = ParamSpec("P")


def maybe_additional_metadata_to_kv(
    additional_metadata: dict[str, str] | JSONSerializableMapping | None
) -> list[str] | None:
    if not additional_metadata:
        return None

    return [f"{k}:{v}" for k, v in additional_metadata.items()]


class BaseRestClient:
    def __init__(self, config: ClientConfig) -> None:
        self.tenant_id = config.tenant_id

        self.client_config = config
        self.api_config = Configuration(
            host=config.server_url,
            access_token=config.token,
        )

        self.api_config.datetime_format = "%Y-%m-%dT%H:%M:%S.%fZ"

    def client(self) -> ApiClient:
        return ApiClient(self.api_config)
